import logging
from typing import Mapping
from typing import Optional

logger = logging.getLogger(__name__)


def _is_deprecated_attribute_name(obj, attr: str) -> bool:
    return (
        attr != "_DEPRECATED_CLASS_ATTRIBUTES"
        and attr in obj._DEPRECATED_CLASS_ATTRIBUTES
    )


def _get_new_attribute_name(obj, old_name: str) -> str:
    new_name = obj._DEPRECATED_CLASS_ATTRIBUTES[old_name]
    logger.warning(
        f"'{old_name}' is deprecated and will be removed in a future version. Use '{new_name}' instead."
    )
    return new_name


class _Meta(type):
    def __new__(
        meta,
        name,
        bases,
        attrs,
        deprecated_class_attributes: Optional[Mapping[str, str]] = None,
        **kw,
    ):
        newcls = super().__new__(meta, name, bases, attrs, **kw)

        deprecated_class_attributes = deprecated_class_attributes or dict()
        if hasattr(newcls, "_DEPRECATED_CLASS_ATTRIBUTES"):
            newcls._DEPRECATED_CLASS_ATTRIBUTES.update(deprecated_class_attributes)
        else:
            newcls._DEPRECATED_CLASS_ATTRIBUTES = deprecated_class_attributes
        return newcls

    def __setattr__(cls, attr: str, value):
        if _is_deprecated_attribute_name(cls, attr):
            attr = _get_new_attribute_name(cls, attr)
        return super().__setattr__(attr, value)

    def __getattr__(cls, attr: str):
        if _is_deprecated_attribute_name(cls, attr):
            attr = _get_new_attribute_name(cls, attr)
            return getattr(cls, attr)
        raise AttributeError(f"'{cls.__name__}' object has no attribute '{attr}'")


class WithDeprecatedClassAttributes(metaclass=_Meta):

    def __setattr__(self, attr: str, value):
        if _is_deprecated_attribute_name(self, attr):
            attr = _get_new_attribute_name(self, attr)
        return super().__setattr__(attr, value)

    def __getattr__(self, attr: str):
        if _is_deprecated_attribute_name(self, attr):
            attr = _get_new_attribute_name(self, attr)
            return getattr(self, attr)
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{attr}'"
        )



from ..import_utils import unavailable_class
from .processor import Id02BaseProcessor

try:
    from id02.acquisition.saxs_preset import ID02DahuProcessingPreset
except ImportError as ex:
    ID02DahuProcessingPreset = unavailable_class(ex)


class Id02SaxsProcessor(Id02BaseProcessor):
    DEFAULT_QUEUE = "saxs"

    def _set_up_preset(self):
        return ID02DahuProcessingPreset()



from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from .utils import directories
from .utils import models
from .workflows import fluoxas_multi_detector
from .workflows import fluoxas_multi_detector_align
from .workflows import fluoxas_nofit_align
from .workflows import fluoxas_single_detector
from .workflows import fluoxas_single_detector_align


def fluoxas_workflow_inputs(
    filenames: Sequence[str],
    output_root_uri: str,
    scan_ranges: Sequence[Tuple[int, int]],
    config_filenames: Sequence[str],
    instrument_name: Optional[str] = None,
    detector_numbers: Optional[Sequence[int]] = None,
    detector_names: Optional[Sequence[str]] = None,
    fast_fitting: bool = True,
    quantification: bool = True,
    diagnostics: bool = True,
    livetime_ref_value: Union[str, int, float, None] = None,
    counter_ref_value: Union[str, int, float, None] = None,
    counter_name: Optional[str] = None,
    energy_name: Optional[str] = None,
    exclude_scans: Optional[Sequence[Sequence[int]]] = None,
    resolution: Optional[Dict[str, Tuple[Union[int, float], str]]] = None,
    do_align: bool = True,
    align_counter: Optional[str] = None,
    align_crop: Optional[bool] = True,
    fast_align_counter_selection: Optional[bool] = False,
    virtual_axes: Optional[Dict[str, str]] = None,
    ignore_axes: Optional[List[str]] = None,
    axis_units: Optional[Dict[str, str]] = None,
) -> Tuple[str, List[dict]]:

    if len(filenames) != len(scan_ranges):
        raise ValueError(
            f"{len(filenames)} scan ranges (first, last) are needed for {len(filenames)} files"
        )

    if not exclude_scans:
        exclude_scans = [()] * len(filenames)

    parameters = models.common_parameters_model(
        detector_numbers,
        detector_names,
        config_filenames,
        energy_name,
        counter_name,
        instrument_name,
        fast_fitting,
        quantification,
        diagnostics,
        livetime_ref_value,
        counter_ref_value,
        True,
        virtual_axes,
        ignore_axes,
        axis_units,
    )

    parameters = models.FluoXasParameters(
        filenames=filenames,
        scan_ranges=scan_ranges,
        exclude_scans=exclude_scans,
        output_root_uri=output_root_uri,
        resolution=resolution,
        align_counter=align_counter,
        fast_align_counter_selection=fast_align_counter_selection,
        align_crop=align_crop,
        **parameters.model_dump(),
    )

    if parameters.no_fit:
        workflow = fluoxas_nofit_align.WORKFLOW
        inputs = fluoxas_nofit_align.workflow_inputs(parameters)
    elif parameters.fit_single_detector:
        if do_align:
            workflow = fluoxas_single_detector_align.WORKFLOW
            inputs = fluoxas_single_detector_align.workflow_inputs(parameters)
        else:
            workflow = fluoxas_single_detector.WORKFLOW
            inputs = fluoxas_single_detector.workflow_inputs(parameters)
    else:
        if parameters.sum_spectra:
            raise ValueError(
                f"Sum before fit is not supported yet for FluoXAS. Provide {len(detector_numbers)} config files."
            )
        if do_align:
            workflow = fluoxas_multi_detector_align.WORKFLOW
            inputs = fluoxas_multi_detector_align.workflow_inputs(parameters)
        else:
            workflow = fluoxas_multi_detector.WORKFLOW
            inputs = fluoxas_multi_detector.workflow_inputs(parameters)

    return workflow, inputs


def fluoxas_paths(
    session: str,
    sample: str,
    datasets: Sequence[str],
    scan_ranges: Sequence[Tuple[int, int]],
    config_filenames: Sequence[str],
    dirname: Optional[str] = None,
    demo: bool = False,
) -> models.FluoXasPaths:
    filenames = [
        directories.raw_directory(session, sample, dataset, demo=demo)
        for dataset in datasets
    ]

    first_dataset = datasets[0]
    first_scan_number = scan_ranges[0][0]

    output_filename = directories.processed_path(
        session,
        sample,
        first_dataset,
        dirname,
        f"{sample}_{first_dataset}.h5",
        demo=demo,
    )
    convert_destination = directories.processed_path(
        session,
        sample,
        first_dataset,
        dirname,
        f"{sample}_{first_dataset}_scan{first_scan_number:04d}.json",
        demo=demo,
    )

    output_root_uri = f"{output_filename}::/{first_scan_number}.1"
    config_filenames = [
        directories.pymca_config_path(session, s, demo=demo) for s in config_filenames
    ]

    workflow_path = directories.accessible_workflow_path(
        session, sample, first_dataset, dirname, demo=demo
    )

    return models.FluoXasPaths(
        filenames=filenames,
        output_root_uri=output_root_uri,
        convert_destination=convert_destination,
        workflow_path=workflow_path,
        config_filenames=config_filenames,
    )



"""Flint-side for Flint RIXS plots"""

import logging
import os

from silx.gui.plot import Plot1D
from silx.io.specfile import SpecFile

from ..flint import capture_errors

logger = logging.getLogger(__name__)


class SpecGenPlotWidget(Plot1D):

    @capture_errors
    def update_plot(
        self, outputfile_sum: str, scan_number: int, max_scans: int
    ) -> None:

        plot_x = "Pixel"
        plot_y = "SPC"

        file_exists = os.path.exists(outputfile_sum)
        logger.info(
            "Update RIXS scan %s (file=%s,file exists=%s)",
            scan_number,
            outputfile_sum,
            file_exists,
        )
        if not file_exists:
            return

        sf = SpecFile(outputfile_sum)
        keys = sf.keys()

        scan_number = int(scan_number)
        my_key = "%d" % scan_number + ".1"

        if my_key not in keys:
            logger.info("Scan %s not found in %s", my_key, outputfile_sum)
            return

        logger.info("%d", keys.index(my_key))
        x = sf[my_key].data_column_by_name(plot_x)
        y = sf[my_key].data_column_by_name(plot_y)
        title = sf.command(keys.index(my_key))
        sf.close()

        legend = str(scan_number)

        self.addCurve(x, y, legend=legend, xlabel=plot_x, ylabel=plot_y)
        self.setActiveCurve(legend)
        self.setGraphTitle(title)

        curves = self.getItems()
        del_items = curves[: max(0, len(curves) - max_scans)]
        for item in del_items:
            self.remove(item)



import logging
import os
from datetime import datetime
from typing import List
from typing import Literal
from typing import Optional
from typing import Type
from typing import Union

from redis import Redis
from redis_om import JsonModel
from redis_om.model.model import NotFoundError

from ..bliss_globals import current_session
from ..import_utils import UnavailableObject
from ..import_utils import is_available
from ..persistent.ndarray import PersistentNdArray
from .compatibility import Field
from .compatibility import get_redis_db_url

logger = logging.getLogger(__name__)

XrpdFieldName = Literal["radial", "azimuthal", "intensity"]


def _create_database_proxy() -> Union[Redis, UnavailableObject]:
    try:
        return Redis.from_url(get_redis_db_url())
    except Exception as ex:
        # We are here because one of these reasons:
        #  - BEACON_HOST is not defined (ValueError).
        #  - BEACON_HOST is wrong (some socket exception).
        #  - Beacon or Redis are down (some socket exception).
        if is_available(current_session):
            # We are in a Bliss session. We should always have access to Redis.
            raise
        # We are not in a Bliss session, most likely the test suite.
        logger.warning(
            "Redis connection not available (%s). If this is not the "
            "test suite there is an issue with BEACON_HOST=%r or Redis itself.",
            ex,
            os.environ.get("BEACON_HOST", None),
        )
        return UnavailableObject(ImportError())


class XrpdPlotInfo(JsonModel, frozen=True):
    scan_name: str
    lima_name: str
    radial_label: str
    azim_label: Optional[str]
    hdf5_url: Optional[str]
    timestamp: datetime = Field(default_factory=datetime.now)
    field_names: List[XrpdFieldName]

    @property
    def legend(self) -> str:
        return f"{self.scan_name} ({self.lima_name})"

    def _get_redis_key(self, field_name: XrpdFieldName) -> str:
        # TODO: remove spaces from redis key but handle existing
        # Redis memory allocated in production
        return f"{self.scan_name}:{self.lima_name}:plot_data:{field_name}"

    def get_data_array(self, field_name: XrpdFieldName) -> PersistentNdArray:
        if field_name not in self.field_names:
            raise KeyError(f"No field {field_name} in this PlotInfo")
        redis_key = self._get_redis_key(field_name)
        logger.debug("XrpdPlotInfo GET %r", redis_key)
        return PersistentNdArray(redis_key)

    def delete_data_arrays(self) -> None:
        for field_name in self.field_names:
            redis_key = self._get_redis_key(field_name)
            logger.debug("XrpdPlotInfo DEL %r", redis_key)
            PersistentNdArray(redis_key).remove()

    @classmethod
    def get(cls, pk) -> Type["XrpdPlotInfo"]:
        try:
            return super().get(pk)
        except NotFoundError:
            # Reraise NotFoundError to get more info in the error message
            raise KeyError(f"PlotInfo not found at {pk}@{cls._meta.database}")

    class Meta:
        database = _create_database_proxy()



import functools
import logging
from typing import Callable
from typing import Optional
from typing import Union

from ..import_utils import unavailable_class
from ..import_utils import unavailable_function
from ..import_utils import unavailable_type

try:
    from bliss.scanning.scan_meta import get_user_scan_meta
except ImportError as ex:
    get_user_scan_meta = unavailable_function(ex)

try:
    from bliss.scanning.scan import Scan as BlissScanType
except ImportError as ex:
    BlissScanType = unavailable_type(ex)

try:
    from bliss.scanning.scan_meta import META_TIMING
except ImportError as ex:
    META_TIMING = unavailable_class(ex)

try:
    from bliss.scanning.scan_state import ScanState
except ImportError as ex:
    ScanState = unavailable_class(ex)

TriggerType = Callable[[BlissScanType], Optional[dict]]

_SCAN_META_CATEGORY = "workflows"

logger = logging.getLogger(__name__)


def enable_processing_trigger(
    processing_id: str,
    trigger: TriggerType,
    timing: Union[str, META_TIMING],
):
    """
    Execute `trigger` when the scan reaches a certain stage (PREPARED by default).

    When `trigger` returns a dictionary it will merged with the dictionary of
    all other trigger functions in the Bliss session and saved in HDF5 under
    the "/x.y/workflows" group. The dictionary follows the convention accepted
    by the `dicttonx` method from the silx library.

    The `processing_id` needs to be unique within a Bliss session.
    """
    if isinstance(timing, str):
        try:
            timing = META_TIMING[timing.upper()]
        except KeyError:
            raise KeyError(
                f"timing must be one of {META_TIMING.__members__.values()}"
            ) from None
    elif not isinstance(timing, META_TIMING):
        raise TypeError(
            f"'timing' must be of type 'str' or 'META_TIMING', got '{type(timing)}'"
        )

    scan_meta_obj = get_user_scan_meta()
    if _SCAN_META_CATEGORY not in scan_meta_obj.categories_names():
        scan_meta_obj.add_categories({_SCAN_META_CATEGORY})
        meta_category = getattr(scan_meta_obj, _SCAN_META_CATEGORY)
        # Make category available for all timings (filter in the `_trigger_on_timing` wrapper)
        for t in META_TIMING.__members__.values():
            meta_category.timing |= t
        meta_category.set("@NX_class", {"@NX_class": "NXcollection"})
    else:
        meta_category = getattr(scan_meta_obj, _SCAN_META_CATEGORY)

    if meta_category.is_set(processing_id):
        raise RuntimeError(
            f"Another processor has already registered the processing id '{processing_id}'"
        )

    trigger_on_state = _trigger_on_timing(timing)(trigger)
    logger.info("Blissoda: trigger %r on %s enabled", processing_id, timing)
    meta_category.set(processing_id, trigger_on_state)


def disable_processing_trigger(processing_id: str):
    scan_meta_obj = get_user_scan_meta()
    if _SCAN_META_CATEGORY not in scan_meta_obj.categories_names():
        return
    meta_category = getattr(scan_meta_obj, _SCAN_META_CATEGORY)
    if meta_category.is_set(processing_id):
        meta_category.remove(processing_id)
        logger.info("Blissoda: trigger %r disabled", processing_id)


def _trigger_on_timing(timing: META_TIMING):
    def _decorator(trigger: TriggerType):
        @functools.wraps(trigger)
        def _wrapper(scan=None):
            if scan is None:
                return
            current_timing = _get_meta_timing(scan)
            if timing & current_timing:
                # The processing was enabled for this timing
                return trigger(scan)

        return _wrapper

    return _decorator


def _get_meta_timing(scan: BlissScanType) -> META_TIMING:
    if scan.state < ScanState.PREPARING:
        return META_TIMING.START
    if scan.state == ScanState.PREPARING:
        return META_TIMING.PREPARED
    return META_TIMING.END



from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from .utils import directories
from .utils import models
from .workflows import xrfmap_multi_detector
from .workflows import xrfmap_multi_detector_sumspectra
from .workflows import xrfmap_single_detector


def xrfmap_workflow_inputs(
    filename: str,
    output_root_uri: str,
    scan_number: int,
    config_filenames: Sequence[str],
    instrument_name: Optional[str] = None,
    detector_numbers: Optional[Sequence[int]] = None,
    detector_names: Optional[Sequence[str]] = None,
    fast_fitting: bool = True,
    quantification: bool = True,
    diagnostics: bool = True,
    livetime_ref_value: Union[str, int, float, None] = None,
    counter_ref_value: Union[str, int, float, None] = None,
    counter_name: Optional[str] = None,
    energy_name: Optional[str] = None,
    resolution: Optional[Dict[str, Tuple[Union[int, float], str]]] = None,
    virtual_axes: Optional[Dict[str, str]] = None,
    ignore_axes: Optional[List[str]] = None,
    axis_units: Optional[Dict[str, str]] = None,
) -> Tuple[str, List[dict]]:
    parameters = models.common_parameters_model(
        detector_numbers,
        detector_names,
        config_filenames,
        energy_name,
        counter_name,
        instrument_name,
        fast_fitting,
        quantification,
        diagnostics,
        livetime_ref_value,
        counter_ref_value,
        False,
        virtual_axes,
        ignore_axes,
        axis_units,
    )

    parameters = models.XrfMapParameters(
        filename=filename,
        scan_number=scan_number,
        output_root_uri=output_root_uri,
        resolution=resolution,
        **parameters.model_dump(),
    )

    if parameters.fit_single_detector:
        if parameters.sum_spectra:
            workflow = xrfmap_multi_detector_sumspectra.WORKFLOW
            inputs = xrfmap_multi_detector_sumspectra.workflow_inputs(parameters)
        else:
            workflow = xrfmap_single_detector.WORKFLOW
            inputs = xrfmap_single_detector.workflow_inputs(parameters)
    else:
        workflow = xrfmap_multi_detector.WORKFLOW
        inputs = xrfmap_multi_detector.workflow_inputs(parameters)

    return workflow, inputs


def xrfmap_paths(
    session: str,
    sample: str,
    dataset: str,
    scan_number: int,
    config_filenames: Sequence[str],
    dirname: Optional[str] = None,
    demo: bool = False,
) -> models.XrfMapPaths:
    filename = directories.raw_directory(session, sample, dataset, demo=demo)

    output_filename = directories.processed_path(
        session, sample, dataset, dirname, f"{sample}_{dataset}.h5", demo=demo
    )
    convert_destination = directories.processed_path(
        session,
        sample,
        dataset,
        dirname,
        f"{sample}_{dataset}_scan{scan_number:04d}.json",
        demo=demo,
    )

    output_root_uri = f"{output_filename}::/{scan_number}.1"
    config_filenames = [
        directories.pymca_config_path(session, s, demo=demo) for s in config_filenames
    ]

    workflow_path = directories.accessible_workflow_path(
        session, sample, dataset, dirname, demo=demo
    )
    return models.XrfMapPaths(
        filename=filename,
        output_root_uri=output_root_uri,
        convert_destination=convert_destination,
        workflow_path=workflow_path,
        config_filenames=config_filenames,
    )



"""User API for XRPD processing on the Bliss repl"""

import json
import logging
import os
import shutil
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from ewoksjob.client import get_future
from ewoksjob.client import submit
from silx.io import h5py_utils

from ..bliss_globals import current_session
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessorWithPlotting
from ..processor import BlissScanType
from ..resources import resource_filename
from ..utils import directories
from .plotter import XrpdPlotter
from .utils import get_integrated_nxdata

logger = logging.getLogger(__name__)


class XrpdProcessor(
    BaseProcessorWithPlotting,
    parameters=[
        ParameterInfo("workflow_with_saving", category="workflows"),
        ParameterInfo("workflow_without_saving", category="workflows"),
        ParameterInfo("queue", category="workflows", deprecated_names=["worker"]),
        ParameterInfo("lima_names", category="data access"),
        ParameterInfo("data_from_memory", category="data access"),
        ParameterInfo("retry_timeout", category="data access"),
        ParameterInfo("retry_period", category="data access"),
        ParameterInfo("flush_period", category="data access"),
        ParameterInfo("lima_url_template", category="data access"),
        ParameterInfo("lima_url_template_args", category="data access"),
        ParameterInfo("prioritize_non_native_h5items", category="data access"),
        ParameterInfo("save_scans_separately", category="data access"),
        ParameterInfo("number_of_scans", category="plotting"),
        ParameterInfo("trigger_from_bliss", category="plotting"),
        ParameterInfo("monitor_name", category="PyFai"),
        ParameterInfo("reference", category="PyFai"),
        ParameterInfo("data_portal_upload", category="data portal"),
    ],
):
    """A class that holds parameters related to online workflow triggering.

    This class must be subclassed to implement:
    - the generation of the pyFAI config file: `get_config_file`
    - the generation of integration options: `get_integration_options`

    These are generally beamline-specific.
    """

    DEFAULT_WORKFLOW: Optional[str] = resource_filename(
        "xrpd", "integrate_scan_with_saving.json"
    )
    DEFAULT_WORKFLOW_NO_SAVE: Optional[str] = resource_filename(
        "xrpd", "integrate_scan_without_saving.json"
    )
    DEFAULT_LIMA_URL_TEMPLATE: Optional[str] = None

    plotter_class = XrpdPlotter

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        enable_plotter: bool = True,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("_plotting_enabled", enable_plotter)
        defaults.setdefault("workflow_with_saving", self.DEFAULT_WORKFLOW)
        defaults.setdefault("workflow_without_saving", self.DEFAULT_WORKFLOW_NO_SAVE)
        defaults.setdefault("lima_names", list())
        defaults.setdefault("data_from_memory", True)
        defaults.setdefault("trigger_from_bliss", True)
        defaults.setdefault("retry_period", 1)
        defaults.setdefault("flush_period", 5)
        defaults.setdefault("lima_url_template", self.DEFAULT_LIMA_URL_TEMPLATE)
        defaults.setdefault("prioritize_non_native_h5items", False)
        defaults.setdefault("save_scans_separately", False)
        defaults.setdefault("data_portal_upload", False)

        self._workflow_subdir = "workflows"

        super().__init__(config=config, defaults=defaults)

    def enable(self, *detectors) -> None:
        self.set_lima_names(*detectors)
        super().enable()

    def set_lima_names(self, *detectors) -> None:
        lima_names = set()
        for detector in detectors:
            lima_names.add(detector.name)
        if lima_names:
            self.lima_names = sorted(lima_names)

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> Optional[dict]:
        return self.on_new_scan_metadata(scan)

    def on_new_scan_metadata(self, scan: BlissScanType) -> Optional[dict]:
        metadata, _ = self._on_new_scan(scan)
        return metadata

    def on_new_scan(self, scan: BlissScanType) -> Optional[Any]:
        _, future = self._on_new_scan(scan)
        return future

    def _on_new_scan(self, scan: BlissScanType) -> Tuple[Optional[dict], Optional[Any]]:
        """Executed at the start of every scan"""
        future = None
        metadata = None
        if not self.scan_requires_processing(scan):
            return metadata, future
        if not self.check_monitor(scan):
            print(
                f"⚠️ WARNING: The workflow cannot run! Monitor '{self.monitor_name}' is not included in the scan. Remove it from the pyfai options."
            )
            return metadata, future

        workflow = self.get_workflow(scan)

        lima_names = self.get_lima_names(scan)

        upload_parameters = None
        if self.data_portal_upload and scan.scan_info.get("save") and lima_names:
            # It is allowed to upload the same processed directory more than once.
            upload_parameters = self._get_workflow_upload_parameters(scan)

        for lima_name in lima_names:
            # Submit arguments
            kwargs = self.get_submit_arguments(scan, lima_name)
            if scan.scan_info.get("save"):
                kwargs["convert_destination"] = self.workflow_destination(
                    scan, lima_name
                )
            if upload_parameters:
                kwargs["upload_parameters"] = upload_parameters

            if self.trigger_from_bliss:
                # Trigger workflow from the current process.
                future = submit(args=(workflow,), kwargs=kwargs, queue=self.queue)
                if self.plotter:
                    self._trigger_plotting(scan, lima_name, future)
                future = get_future(future.uuid)
            else:
                # Save metadata to Redis (is saved in the raw data file as well).
                # Workflows may be triggered by an external process as a result.
                kwargs["workflow"] = workflow
                metadata = dict()
                metadata[f"process_{lima_name}"] = {
                    "@NX_class": "NXprocess",
                    "program": "ewoks",
                    "configuration": {
                        "@NX_class": "NXnote",
                        "type": "application/json",
                        "data": json.dumps(kwargs),
                    },
                }

        return metadata, future

    def _data_to_plot_url(self, scan: BlissScanType, lima_name: str):
        if not scan.scan_info.get("save"):
            return None

        output_url = self.online_output_url(scan, lima_name)
        return f"{output_url}/{lima_name}_integrate/integrated"

    def _trigger_plotting(self, scan: BlissScanType, lima_name: str, future) -> None:
        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")

        scan_name = os.path.splitext(os.path.basename(filename))[0]
        scan_name = f"{scan_name}: {scan_nb}.1 {scan.name}"

        output_url = self._data_to_plot_url(scan, lima_name)

        if self.plotter:
            self.plotter.handle_workflow_result(
                future,
                lima_name,
                scan_name,
                output_url=output_url,
                retry_timeout=self.retry_timeout,
                retry_period=self.retry_period,
            )

    def get_submit_arguments(self, scan: BlissScanType, lima_name) -> dict:
        return {
            "inputs": self.get_inputs(scan, lima_name),
            "outputs": [{"all": False}],
        }

    def _get_workflow(self, scan: BlissScanType) -> Optional[str]:
        """Get the workflow filename for the scan"""
        if scan.scan_info.get("save"):
            return self.workflow_with_saving
        else:
            return self.workflow_without_saving

    def _get_default_workflow(self, scan: BlissScanType) -> Optional[str]:
        """Get the workflow filename for the scan"""
        if scan.scan_info.get("save"):
            return self.DEFAULT_WORKFLOW
        else:
            return self.DEFAULT_WORKFLOW_NO_SAVE

    def _set_workflow(self, scan: BlissScanType, filename) -> None:
        """Set the workflow filename for the scan"""
        if scan.scan_info.get("save"):
            self.workflow_with_saving = filename
        else:
            self.workflow_without_saving = filename

    def get_filename(self, scan: BlissScanType) -> str:
        filename = scan.scan_info.get("filename")
        if filename:
            return filename
        return current_session.scan_saving.filename

    def get_workflow(self, scan: BlissScanType) -> Optional[str]:
        """Get the workflow to execute for the scan and ensure it is located
        in the proposal directory for user reference and worker accessibility.
        """
        src_file = self._get_workflow(scan)
        if src_file is None:
            return
        if not os.path.isfile(src_file):
            src_file = self._get_default_workflow(scan)

        dataset_filename = self.get_filename(scan)
        workflow_directory = self._get_workflows_dir(dataset_filename)
        dst_file = os.path.join(workflow_directory, os.path.basename(src_file))
        if src_file != dst_file:
            self._set_workflow(scan, dst_file)

        if not os.path.exists(dst_file):
            os.makedirs(workflow_directory, exist_ok=True)
            shutil.copyfile(src_file, dst_file)

        return dst_file

    def _get_workflows_dir(self, dataset_filename: str) -> str:
        return directories.get_workflows_dir(dataset_filename)

    def get_lima_names(self, scan: BlissScanType) -> List[str]:
        channels = scan.scan_info.get("channels", dict())
        return sorted(
            (
                lima_name
                for lima_name in self.lima_names
                if f"{lima_name}:image" in channels
            )
        )

    def check_monitor(self, scan: BlissScanType) -> bool:
        monitor_name = self.monitor_name
        if not monitor_name:
            return True
        channels = scan.scan_info.get("channels", dict())
        return any(monitor_name in name for name in channels)

    def scan_requires_processing(self, scan: BlissScanType) -> bool:
        return bool(self.get_lima_names(scan)) and self._get_workflow(scan) is not None

    def scan_processed_directory(self, scan: BlissScanType) -> str:
        return directories.get_dataset_processed_dir(self.get_filename(scan))

    def workflow_destination(self, scan: BlissScanType, lima_name: str) -> str:
        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        root = self.scan_processed_directory(scan)
        stem = os.path.splitext(os.path.basename(filename))[0]
        lima_sep = len(self.lima_names) > 1
        if lima_sep:
            basename = f"{stem}_{lima_name}_{scan_nb:04d}.json"
        else:
            basename = f"{stem}_{scan_nb:04d}.json"
        return os.path.join(root, self._workflow_subdir, basename)

    def master_output_filename(self, scan: BlissScanType) -> str:
        """Filename which can be used to inspect the results after the processing."""
        filename = self.get_filename(scan)
        root = self.scan_processed_directory(scan)
        basename = os.path.basename(filename)
        return os.path.join(root, basename)

    def external_output_filename(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[str]:
        """External filename in which the online processing saves the results.
        When `None` the online processing saves directly in the master file.
        """
        lima_sep = len(self.lima_names) > 1
        scan_sep = self.save_scans_separately
        if not lima_sep and not scan_sep:
            return  # save directly in the master file

        filename = self.get_filename(scan)
        root = self.scan_processed_directory(scan)
        basename = os.path.basename(filename)

        stem, ext = os.path.splitext(basename)
        stem_parts = [stem]
        if lima_sep:
            stem_parts.append(lima_name)
        if scan_sep:
            scan_nb = scan.scan_info.get("scan_nb")
            stem_parts.append(f"{scan_nb:04d}")
        basename = "_".join(stem_parts)

        return os.path.join(root, f"{basename}{ext}")

    def online_output_filename(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[str]:
        """Filename which can be used to inspect the results during and after the processing."""
        filename = self.external_output_filename(scan, lima_name)
        if not filename:
            return self.master_output_filename(scan)
        return filename

    def master_output_url(self, scan: BlissScanType) -> str:
        """URL which can be used to inspect the results after the processing."""
        scan_nb = scan.scan_info.get("scan_nb")
        filename = self.master_output_filename(scan)
        return f"{filename}::/{scan_nb}.1"

    def external_output_url(self, scan: BlissScanType, lima_name: str) -> Optional[str]:
        """External URL in which the online processing saves the results.
        When `None` the online processing saves directly in the master URL.
        """
        scan_nb = scan.scan_info.get("scan_nb")
        filename = self.external_output_filename(scan, lima_name)
        if filename:
            return f"{filename}::/{scan_nb}.1"

    @h5py_utils.retry()
    def get_data_keys(self, scan: BlissScanType, lima_name: str):
        filename = self.online_output_filename(scan, lima_name)

        if not filename:
            raise KeyError(
                f"Could not find a processed filename for this scan and lima name: {lima_name}"
            )

        with h5py_utils.File(filename, mode="r") as root:
            nxdata = get_integrated_nxdata(root, scan)

            return [f"{lima_name}:{key}" for key in nxdata.keys()]

    @h5py_utils.retry()
    def get_data(self, scan: BlissScanType, channel: str, idx=tuple()):
        """
        Get processed data for a scan. Tries to mirror the existing `scan.get_data` for raw data.
        A list of available channels for a given lima name can be retrieved using `get_data_keys`.

        Ex:
            >>> scan = loopscan(5, 0.1, difflab6)
            >>> xrpd_processor.get_data_keys(scan, 'difflab6')
            ['difflab6:intensity', 'difflab6:points', 'difflab6:q']
            >>> xrpd_processor.get_data(scan, 'difflab6:intensity')
            array([...], dtype=float32)
        """
        lima_name, field = channel.split(":")
        filename = self.online_output_filename(scan, lima_name)

        if not filename:
            raise KeyError(
                f"Could not find a processed filename for this scan and lima name: {lima_name}"
            )

        with h5py_utils.File(filename, mode="r") as root:
            nxdata = get_integrated_nxdata(root, scan)

            if field not in nxdata:
                raise KeyError(
                    f"{channel} is not a channel of this processing. Possible channels: {[f'{lima_name}:{key}' for key in nxdata.keys()]}."
                )

            return nxdata[field][idx]

    def online_output_url(self, scan: BlissScanType, lima_name: str) -> str:
        """URL which can be used to inspect the results during and after the processing."""
        scan_nb = scan.scan_info.get("scan_nb")
        filename = self.online_output_filename(scan, lima_name)
        return f"{filename}::/{scan_nb}.1"

    def get_inputs(self, scan: BlissScanType, lima_name: str) -> List[dict]:
        inputs = self.get_config_inputs(scan, lima_name)
        inputs += self.get_integrate_inputs(scan, lima_name, "IntegrateBlissScan")
        inputs += self.get_integrate_inputs(
            scan, lima_name, "IntegrateBlissScanWithoutSaving"
        )
        inputs += self.get_save_inputs(scan, lima_name, "SaveNexusIntegrated")
        inputs += self.get_integrate_list_inputs(scan, lima_name)
        inputs += self.get_integrate_1d_inputs(scan, lima_name)
        return inputs

    def _get_common_integrate_inputs(self, task_identifier: str) -> List[dict]:
        """Inputs common to all :math:`ewoksxrpd` integration tasks."""
        inputs = [
            {
                "task_identifier": task_identifier,
                "name": "maximum_persistent_workers",
                "value": len(self.lima_names),
            }
        ]
        inputs += self._get_data_access_inputs(task_identifier)
        return inputs

    def _get_data_access_inputs(self, task_identifier: str) -> List[dict]:
        """Inputs common to all :math:`ewoksxrpd` tasks deriving from `TaskWithDataAccess`."""
        inputs = [
            {
                "task_identifier": task_identifier,
                "name": "retry_timeout",
                "value": self.retry_timeout,
            },
            {
                "task_identifier": task_identifier,
                "name": "retry_period",
                "value": self.retry_period,
            },
            {
                "task_identifier": task_identifier,
                "name": "lima_url_template",
                "value": self.lima_url_template,
            },
            {
                "task_identifier": task_identifier,
                "name": "prioritize_non_native_h5items",
                "value": self.prioritize_non_native_h5items,
            },
        ]
        if self.lima_url_template_args:
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "lima_url_template_args",
                    "value": dict(self.lima_url_template_args),
                }
            )
        return inputs

    def get_integrate_inputs(
        self, scan: BlissScanType, lima_name: str, task_identifier: str
    ) -> List[dict]:
        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        inputs = [
            {
                "task_identifier": task_identifier,
                "name": "filename",
                "value": filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "output_filename",
                "value": self.master_output_filename(scan),
            },
            {
                "task_identifier": task_identifier,
                "name": "scan",
                "value": scan_nb,
            },
            {
                "task_identifier": task_identifier,
                "name": "detector_name",
                "value": lima_name,
            },
            {
                "task_identifier": task_identifier,
                "name": "monitor_name",
                "value": self.monitor_name,
            },
            {
                "task_identifier": task_identifier,
                "name": "reference",
                "value": self.reference,
            },
            {
                "task_identifier": task_identifier,
                "name": "flush_period",
                "value": self.flush_period,
            },
        ]
        external_output_filename = self.external_output_filename(scan, lima_name)
        if external_output_filename:
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "external_output_filename",
                    "value": self.external_output_filename(scan, lima_name),
                }
            )
        if self.data_from_memory or not scan.writer.device_saving_enabled():
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "scan_memory_url",
                    "value": _get_scan_memory_url(scan),
                }
            )
        inputs += self._get_common_integrate_inputs(task_identifier)
        return inputs

    def get_integrate_list_inputs(
        self, scan: BlissScanType, lima_name: str
    ) -> List[dict]:
        task_identifier = "Integrate1DList"
        # TODO: monitor_name needs to be provided to the workflow
        scan_nb = scan.scan_info.get("scan_nb")
        inputs = [
            {
                "task_identifier": task_identifier,
                "name": "output_file",
                "value": self.master_output_filename(scan),
            },
            {
                "task_identifier": task_identifier,
                "name": "entry_name",
                "value": f"{scan_nb}.1",
            },
            {
                "task_identifier": task_identifier,
                "name": "reference",
                "value": self.reference,
            },
            {
                "task_identifier": task_identifier,
                "name": "flush_period",
                "value": self.flush_period,
            },
        ]
        inputs += self._get_common_integrate_inputs(task_identifier)
        return inputs

    def get_integrate_1d_inputs(
        self, scan: BlissScanType, lima_name: str
    ) -> List[dict]:
        task_identifier = "Integrate1D"
        inputs = [
            {
                "task_identifier": task_identifier,
                "name": "detector_name",
                "value": lima_name,
            },
            {
                "task_identifier": task_identifier,
                "name": "monitor",
                "value": self.monitor_name,
            },
            {
                "task_identifier": task_identifier,
                "name": "reference",
                "value": self.reference,
            },
        ]
        inputs += self._get_common_integrate_inputs(task_identifier)
        return inputs

    def get_save_inputs(
        self, scan: BlissScanType, lima_name: str, task_identifier: str
    ) -> List[dict]:
        inputs = self._get_data_access_inputs(task_identifier)
        inputs.append(
            {
                "task_identifier": task_identifier,
                "name": "url",
                "value": self.master_output_url(scan),
            }
        )
        external_url = self.external_output_url(scan, lima_name)
        if external_url:
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "external_url",
                    "value": external_url,
                }
            )
        return inputs

    def get_config_inputs(self, scan: BlissScanType, lima_name: str) -> List[dict]:
        inputs = []
        filename = self.get_config_filename(lima_name)
        if filename:
            inputs.append(
                {
                    "task_identifier": "PyFaiConfig",
                    "name": "filename",
                    "value": filename,
                }
            )
        integration_options = self.get_integration_options(scan, lima_name)
        if integration_options:
            try:
                integration_options = integration_options.to_dict()
            except AttributeError:
                pass
            inputs.append(
                {
                    "task_identifier": "PyFaiConfig",
                    "name": "integration_options",
                    "value": integration_options,
                }
            )
        return inputs

    def _get_workflow_upload_parameters(self, scan: BlissScanType) -> dict:
        raw_directory = os.path.dirname(self.get_filename(scan))
        processed_directory = self.scan_processed_directory(scan)

        scan_saving = current_session.scan_saving

        metadata = {"Sample_name": scan_saving.dataset["Sample_name"]}

        return {
            "beamline": scan_saving.beamline,
            "proposal": scan_saving.proposal_name,
            "dataset": "integrate",  # processed dataset name
            "path": processed_directory,
            "raw": [raw_directory],
            "metadata": metadata,
        }

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        raise NotImplementedError

    def get_integration_options(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[dict]:
        raise NotImplementedError


def _get_scan_memory_url(scan) -> str:
    try:
        return scan._scan_data.key
    except AttributeError:
        # bliss < 2.0.0
        return f"{scan.root_node.db_name}:{scan._node_name}"



from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ..bliss_globals import current_session
from ..persistent.parameters import ParameterInfo
from ..processor import BlissScanType
from ..resources import resource_filename
from ..utils import validators
from ..xrpd.processor import XrpdProcessor
from .utils import get_data_url
from .utils import get_positions
from .utils import is_multigeometry


class Id06XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config", category="PyFai", validator=validators.is_file),
        ParameterInfo("integration_options", category="PyFai"),
        ParameterInfo("goniometer_file", category="PyFai"),
        ParameterInfo("multigeometry_workflow", category="workflows"),
    ],
):
    MULTIGEOMETRY_WORKFLOW = resource_filename(
        "id06", "integrate_2dmultigeometry_with_saving.json"
    )

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("multigeometry_enabled", False)
        defaults.setdefault("step", 10)
        defaults.setdefault("multigeometry_workflow", self.MULTIGEOMETRY_WORKFLOW)

        super().__init__(config=config, defaults=defaults)

    def _get_workflow(self, scan: BlissScanType) -> str:
        if is_multigeometry(scan):
            return self.multigeometry_workflow
        elif scan.scan_info["save"]:
            return self.workflow_with_saving
        else:
            return self.workflow_without_saving

    def _set_workflow(self, scan: BlissScanType, filename: str) -> None:
        if is_multigeometry(scan):
            self.multigeometry_workflow = filename
        elif scan.scan_info.get("save"):
            self.workflow_with_saving = filename
        else:
            self.workflow_without_saving = filename

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.pyfai_config

    def get_integration_options(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[dict]:
        integration_options = self.integration_options
        if integration_options:
            return integration_options.to_dict()
        return None

    def get_inputs(self, scan, lima_name: str) -> List[dict]:
        inputs = super().get_inputs(scan, lima_name)
        if is_multigeometry(scan):
            inputs += self._get_multigeometry_inputs(scan, lima_name)
        return inputs

    def _get_multigeometry_inputs(
        self, scan: BlissScanType, lima_name: str
    ) -> List[dict]:
        task_identifier = "Integrate2DMultiGeometry"
        multigeometry_inputs = [
            {
                "task_identifier": task_identifier,
                "name": "goniometer_file",
                "value": self._get_parameter("goniometer_file"),
            },
            {
                "task_identifier": task_identifier,
                "name": "positions",
                "value": self._get_positions(),
            },
            {
                "task_identifier": task_identifier,
                "name": "images",
                "value": get_data_url(scan, lima_name),
            },
        ]
        return multigeometry_inputs

    def _get_positions(self) -> List[float]:
        fscan_parameters = current_session.setup_globals.fscan.pars
        return get_positions(
            start=fscan_parameters.start_pos,
            step=fscan_parameters.step_size,
            npts=fscan_parameters.npoints,
        )



import json
import logging
import os
import pathlib
from typing import Any
from typing import Dict
from typing import Optional

from ewoksjob.client import submit

from ..bliss_globals import current_session
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..resources import resource_filename
from ..utils.directories import get_dataset_processed_dir

logger = logging.getLogger(__name__)


class ID32Hdf5ToSpecProcessor(
    BaseProcessor,
    parameters=[
        ParameterInfo("save_single_scans", category="workflows"),
    ],
):
    QUEUE = "lid32xmcd2"
    WORKFLOW_FILENAME = "convert_hdf5_to_spec.json"

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("trigger_at", "END")
        defaults.setdefault("save_single_scans", True)

        super().__init__(config=config, defaults=defaults)

    def _get_workflow(self) -> dict:
        with open(resource_filename("id32", self.WORKFLOW_FILENAME), "r") as wf:
            return json.load(wf)

    def _get_workflow_inputs(self, scan) -> list:
        return [
            {
                "name": "scan_numbers",
                "value": [scan.scan_info.get("scan_nb")],
            },
            {
                "name": "save_single_scans",
                "value": self.save_single_scans,
            },
            {
                "name": "input_file",
                "value": self._get_scan_filename(scan),
            },
            {
                "name": "output_path",
                "value": self._get_scan_processed_directory(scan),
            },
        ]

    def _get_scan_filename(self, scan) -> str:
        filename = scan.scan_info.get("filename")
        if filename:
            return filename
        return current_session.scan_saving.filename

    def _get_scan_processed_directory(self, scan) -> str:
        return get_dataset_processed_dir(self._get_scan_filename(scan))

    def _get_workflow_destination(self, scan) -> str:
        """Builds the path where the workflow JSON will be saved."""
        filename = self._get_scan_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        root = self._get_scan_processed_directory(scan)
        stem = os.path.splitext(os.path.basename(filename))[0]
        wf_path = os.path.join(root, "workflows")
        pathlib.Path(wf_path).mkdir(parents=True, exist_ok=True)
        basename = f"{stem}_{scan_nb:04d}_make_specfile.json"
        return os.path.join(wf_path, basename)

    def _trigger_workflow_on_new_scan(self, scan) -> None:
        if not scan.scan_info["save"]:
            return

        workflow = self._get_workflow()
        inputs = self._get_workflow_inputs(scan)
        kwargs = {"inputs": inputs, "outputs": [{"all": False}]}
        kwargs["convert_destination"] = self._get_workflow_destination(scan)

        _ = submit(args=(workflow,), kwargs=kwargs, queue=self.QUEUE)



from __future__ import annotations

import logging
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import Optional

from .automation import BlissAutomationObject
from .flint.plotter import BasePlotter
from .import_utils import unavailable_type
from .persistent.parameters import ParameterInfo
from .persistent.parameters import autocomplete_property
from .utils import trigger

try:
    from bliss.scanning.scan import Scan as BlissScanType
except ImportError as ex:
    BlissScanType = unavailable_type(ex)

logger = logging.getLogger(__name__)


class BaseProcessor(
    BlissAutomationObject,
    parameters=[
        ParameterInfo("_enabled"),
        ParameterInfo("trigger_at", category="workflows"),
    ],
):
    """Enable and disable workflow triggering on new scans."""

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("_enabled", False)
        defaults.setdefault("trigger_at", "PREPARED")  # START, PREPARED, END

        super().__init__(config=config, defaults=defaults)

        if self._HAS_BLISS:
            if self._enabled:
                self._register_workflow_trigger()
            else:
                self._unregister_workflow_trigger()

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        status = categories.setdefault("status", {})
        status["Enabled"] = self._enabled
        return categories

    @autocomplete_property
    def trigger_at(self) -> Optional[str]:
        return self._get_parameter("trigger_at")

    @trigger_at.setter
    def trigger_at(self, value: str):
        self._set_parameter("trigger_at", value)
        if self._enabled:
            self.disable()
            self.enable()

    def enable(self):
        if self._enabled:
            return
        self._register_workflow_trigger()
        self._enabled = True

    def disable(self):
        if not self._enabled:
            return
        self._unregister_workflow_trigger()
        self._enabled = False

    def _register_workflow_trigger(self):
        trigger.enable_processing_trigger(
            self._processing_id(),
            self._trigger_workflow_on_new_scan,
            self.trigger_at,
        )

    def _unregister_workflow_trigger(self):
        trigger.disable_processing_trigger(
            self._processing_id(),
        )

    @classmethod
    def _processing_id(cls):
        return cls.__name__

    @abstractmethod
    def _trigger_workflow_on_new_scan(self, scan: BlissScanType) -> Optional[dict]:
        pass


class BaseProcessorWithPlotting(
    BaseProcessor,
    parameters=[
        ParameterInfo("_plotting_enabled"),
        ParameterInfo("number_of_scans", category="plotting"),
    ],
):
    """Enable and disable workflow triggering and Flint plotting on new scans."""

    plotter_class = BasePlotter

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("_plotting_enabled", True)
        defaults.setdefault("number_of_scans", 4)

        super().__init__(config=config, defaults=defaults)

        if self._plotting_enabled:
            self._plotter = self.plotter_class(self.number_of_scans)
            if self._HAS_BLISS:
                self._plotter.replot()
        else:
            self._plotter = None

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        status = categories.setdefault("status", {})
        status["Plots in Flint"] = self._plotter is not None
        if self._plotter:
            categories["status"]["Plotting tasks"] = self._plotter.purge_tasks()
        return categories

    @property
    def plotter(self) -> Optional[plotter_class]:
        return self._plotter

    def enable_plotting(self):
        if self._plotter is not None:
            return
        self._plotter = self.plotter_class(self.number_of_scans)
        self._plotter.replot()
        self._plotting_enabled = True

    def disable_plotting(self):
        if self._plotter is None:
            return
        self.stop_plotting_tasks()
        self._plotter = None
        self._plotting_enabled = False

    def clear_plots(self) -> None:
        if self._plotter:
            return self._plotter.clear()
        else:
            print("Plotting is disabled")

    def replot(self) -> None:
        if self._plotter:
            return self._plotter.replot()
        else:
            print("Plotting is disabled")

    def purge_plotting_tasks(self) -> int:
        if self._plotter:
            return self._plotter.purge_tasks()
        else:
            print("Plotting is disabled")
            return 0

    def stop_plotting_tasks(self) -> int:
        if self._plotter:
            return self._plotter.kill_tasks()
        else:
            print("Plotting is disabled")
            return 0

    @autocomplete_property
    def number_of_scans(self) -> int:
        return self._get_parameter("number_of_scans")

    @number_of_scans.setter
    def number_of_scans(self, value: int):
        self._plotter.number_of_scans = value
        self._set_parameter("number_of_scans", self._plotter.number_of_scans)



"""Bliss-side client for Flint EXAFS plots"""

from typing import List
from typing import Optional
from typing import Tuple

from ..flint import BasePlot
from .types import XasSubscanData


class ExafsPlot(BasePlot):
    WIDGET = "blissoda.exafs.widgets.ExafsWidget"

    def clear(self) -> None:
        self.submit("clear")

    def remove_scan(self, legend: str) -> None:
        self.submit("remove_scan", legend)

    def update_scan(
        self,
        legend: str,
        data: XasSubscanData,
        color: Optional[Tuple[int, int, int]] = None,
    ) -> None:
        self.submit("update_scan", legend, data, color=color)

    def get_scans(self) -> List[str]:
        return self.submit("get_scans")



from __future__ import annotations

import logging
from pathlib import Path
from typing import Tuple

import numpy

from ..bliss_globals import current_session
from ..bliss_globals import setup_globals
from ..import_utils import unavailable_function
from ..utils import directories

try:
    from scipy.io import savemat
except ImportError as ex:
    savemat = unavailable_function(ex)


_logger = logging.getLogger(__name__)


def ensure_difflab6_id31_flats() -> Tuple[str, str]:
    """Create dummy flat field files for ID31 FlatFieldFromEnergy task"""
    difflab6_image_shape = (
        setup_globals.difflab6.image.height,
        setup_globals.difflab6.image.width,
    )

    processed_dir = directories.get_processed_dir(current_session.scan_saving.filename)

    newflat_path = Path(processed_dir, "config", "flats.mat")
    if not newflat_path.is_file():
        _logger.info(f"Create ID31 demo newflat file: {str(newflat_path)}")
        newflat_path.parent.mkdir(parents=True, exist_ok=True)
        savemat(
            newflat_path,
            {
                "E": numpy.array([65, 105], dtype=numpy.uint8),
                "F": numpy.ones(difflab6_image_shape + (2,), dtype=numpy.float64),
            },
        )

    oldflat_path = Path(processed_dir, "config", "flats_old.mat")
    if not oldflat_path.is_file():
        _logger.info(f"Create ID31 demo oldflat file: {str(oldflat_path)}")
        oldflat_path.parent.mkdir(parents=True, exist_ok=True)
        savemat(
            oldflat_path,
            {
                "Eold": numpy.array([15.77, 74.96], dtype=numpy.float64),
                "Fold": numpy.ones(difflab6_image_shape + (2,), dtype=numpy.float64),
            },
        )

    return str(newflat_path), str(oldflat_path)



from typing import List

from ..utils import defaults
from ..utils.models import FluoXasParameters

WORKFLOW = "fluoxas_multi_detector"


def workflow_inputs(parameters: FluoXasParameters) -> List[dict]:
    inputs = [
        {
            "name": "filenames",
            "value": parameters.filenames,
            "task_identifier": "PickScans",
        },
        {
            "name": "scan_ranges",
            "value": parameters.scan_ranges,
            "task_identifier": "PickScans",
        },
        {
            "name": "exclude_scans",
            "value": parameters.exclude_scans,
            "task_identifier": "PickScans",
        },
        {
            "name": "output_root_uri",
            "value": parameters.output_root_uri,
            "task_identifier": parameters.fit_identifier,
        },
        {
            "name": "configs",
            "value": parameters.config_filenames,
            "task_identifier": parameters.fit_identifier,
        },
        {
            "name": "fast_fitting",
            "value": parameters.fast_fitting,
            "task_identifier": parameters.fit_identifier,
        },
        {
            "name": "quantification",
            "value": parameters.quantification,
            "task_identifier": parameters.fit_identifier,
        },
        {
            "name": "diagnostics",
            "value": parameters.diagnostics,
            "task_identifier": parameters.fit_identifier,
        },
        {
            "name": "detector_names",
            "value": parameters.detector_names,
            "task_identifier": parameters.fit_identifier,
        },
        {
            "name": "energy_name",
            "value": parameters.energy_name,
            "task_identifier": parameters.fit_identifier,
        },
        {
            "name": "detector_normalization_template",
            "value": parameters.detector_normalization_template,
            "task_identifier": "SumXrfResultsStack",
        },
        {
            "name": "counter_name",
            "value": parameters.counter_name,
            "task_identifier": parameters.norm_identifier,
        },
        {
            "name": "counter_normalization_template",
            "value": parameters.counter_normalization_template,
            "task_identifier": parameters.norm_identifier,
        },
        {
            "name": "axes_units",
            "value": parameters.axis_units
            or defaults.AXES_UNITS[parameters.instrument_name],
            "task_identifier": "RegridXrfResultsStack",
        },
        {
            "name": "ignore_positioners",
            "value": (
                parameters.ignore_axes
                if parameters.ignore_axes is not None
                else defaults.IGNORE_AXES[parameters.instrument_name]
            ),
            "task_identifier": "RegridXrfResultsStack",
        },
        {
            "name": "resolution",
            "value": parameters.resolution,
            "task_identifier": "RegridXrfResultsStack",
        },
    ]

    return inputs



from ..import_utils import unavailable_module
from .access import WithFlintAccess

try:
    import gevent
except ImportError as ex:
    gevent = unavailable_module(ex)


class BasePlotter(WithFlintAccess):
    def __init__(self, max_plots) -> None:
        super().__init__()
        self._max_plots = max_plots
        self._tasks = []

    @property
    def number_of_scans(self):
        """Maximum number of scans to be plotted"""
        return self._max_plots

    @number_of_scans.setter
    def number_of_scans(self, value):
        self._max_plots = max(value, 0)

    def handle_workflow_result(self):
        raise NotImplementedError()

    def _spawn(self, *args, **kw):
        task = gevent.spawn(*args, **kw)
        self._tasks.append(task)
        self.purge_tasks()

    def purge_tasks(self) -> int:
        """Remove references to tasks that have finished."""
        self._tasks = [t for t in self._tasks if t]
        return len(self._tasks)

    def kill_tasks(self) -> int:
        """Kill all tasks."""
        gevent.killall(self._tasks)
        return self.purge_tasks()

    def replot(self, **retry_options) -> None:
        raise NotImplementedError()



from __future__ import annotations

import json
import logging
import os
from typing import Any

from ..import_utils import unavailable_module

try:
    import gevent
except ImportError as ex:
    gevent = unavailable_module(ex)
import numpy as np
from ewoksjob.client import get_future
from ewoksjob.client import submit

from blissoda.resources import resource_filename

from ..bliss_globals import current_session  # type: ignore
from ..flint.access import WithFlintAccess
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType  # type: ignore
from .plots import AggregatedScansPlot

logger = logging.getLogger(__name__)


class ScansAggregator(
    BaseProcessor,
    parameters=[
        ParameterInfo("workflow", category="workflows"),
        ParameterInfo("data_mappings", category="workflows"),
        ParameterInfo("filename", category="workflows"),
        ParameterInfo("start_scan_id", category="workflows"),
        ParameterInfo("stop_scan_id", category="workflows"),
        ParameterInfo("aggregation_mode", category="workflows"),
        ParameterInfo("use_daxs", category="workflows"),
    ],
):
    DEFAULT_PARAMETERS = {
        "workflow": resource_filename("id26", "scans_aggregator.json"),
        "data_mappings": {
            "x": ".1/measurement/elapsed_time",
            "signal": ".1/measurement/sim_gaussian_1",
        },
        "filename": None,
        "start_scan_id": None,
        "stop_scan_id": None,
        "aggregation_mode": "fraction of sums",
        "use_daxs": True,
        "trigger_at": "END",
        "_enabled": False,
        "_clear_plot": True,
    }

    def __init__(
        self,
        config: dict[str, Any] | None = None,
        defaults: dict[str, Any] | None = None,
        **deprecated_defaults: dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        for parameter, value in self.DEFAULT_PARAMETERS.items():
            defaults.setdefault(parameter, value)

        super().__init__(config=config, defaults=defaults)

        self.handler = AggregatedScansHandler()

    def reset(self):
        for parameter, value in self.DEFAULT_PARAMETERS.items():
            if parameter == "_enabled" and self._enabled:
                self.disable()  # Only disable if currently enabled.
            elif parameter == "data_mappings":
                continue  # Do not reset data mappings
            else:
                setattr(self, parameter, value)
        self.handler.clear()

    def _get_current_scan_id(self) -> None | int:
        if current_session is None:
            raise ValueError("No current session available to get the scan from.")
        try:
            scan = current_session.scans[-1]
        except (AttributeError, IndexError):
            raise ValueError("No scans found in the current session.")

        scan_id = scan.scan_info["scan_nb"]
        if not isinstance(scan_id, int):
            raise ValueError("The id of the current scan is not a valid integer.")
        return scan_id

    def _get_start_scan_id(self) -> None | int:
        return self._get_parameter("start_scan_id")  # type: ignore

    def _set_start_scan_id(self, value: None | int) -> None:
        if value is None:
            self._set_parameter("start_scan_id", None)
            return
        if value < 0:
            current_scan_id = self._get_current_scan_id()
            if not isinstance(current_scan_id, int):
                raise ValueError(
                    "The id of the current scan must be be a valid integer before"
                    " using negative start_scan_id."
                )
            value = value + current_scan_id + 1
        if value == 0:
            raise ValueError("The start_scan_id cannot be zero.")
        self._set_parameter("start_scan_id", value)

    def _get_stop_scan_id(self) -> None | int:
        return self._get_parameter("stop_scan_id")  # type: ignore

    def _set_stop_scan_id(self, value: None | int) -> None:
        if value is None:
            self._set_parameter("stop_scan_id", None)
            return
        if value < 0:
            current_scan_id = self._get_current_scan_id()
            if not isinstance(current_scan_id, int):
                raise ValueError(
                    "The id of the current scan must be be a valid integer before"
                    " using negative stop_scan_id."
                )
            value = value + current_scan_id + 1
        if value == 0:
            raise ValueError("The stop_scan_id cannot be zero.")
        self._set_parameter("stop_scan_id", value)

    def _get_workflow(self) -> dict:
        with open(self.workflow) as wf:
            return json.load(wf)

    def _get_workflow_inputs(self, scan) -> list:
        if scan is not None:
            current_scan_id = self._get_current_scan_id()
            if self._get_start_scan_id() is None:
                self._set_start_scan_id(current_scan_id)
            self._set_stop_scan_id(current_scan_id)

        return [
            {
                "name": "data_mappings",
                "value": dict(self.data_mappings),
            },
            {
                "name": "filename",
                "value": self._get_filename(scan),
            },
            {
                "name": "start_scan_id",
                "value": self._get_parameter("start_scan_id"),
            },
            {
                "name": "stop_scan_id",
                "value": self._get_parameter("stop_scan_id"),
            },
            {
                "name": "aggregation_mode",
                "value": self._get_parameter("aggregation_mode"),
            },
            {
                "name": "use_daxs",
                "value": self._get_parameter("use_daxs"),
            },
        ]

    def _get_filename(self, scan=None) -> str:
        if self.filename is not None:
            return self.filename

        if scan is not None:
            filename = scan.scan_info.get("filename")
            if filename is not None:
                self.filename = filename
                return filename
            else:
                raise ValueError("Scan has no filename.")

        if current_session is None:
            raise ValueError("No current session available to get the filename from.")

        return current_session.scan_saving.filename

    def _get_processed_data_path(self) -> str:
        if self.filename is None:
            raise ValueError("Failed to get processed data path.")

        stem = os.path.splitext(os.path.basename(self.filename))[0]

        if current_session is not None:
            scan_saving = current_session.scan_saving
            root = os.path.join(
                scan_saving.base_path,
                scan_saving.proposal_dirname,
                scan_saving.beamline,
                scan_saving.proposal_session_name,
                "PROCESSED_DATA",
            )
        else:
            root = ""

        start_scan_id = self._get_start_scan_id()
        stop_scan_id = self._get_stop_scan_id()

        return os.path.join(
            root, f"{stem}_aggregated_scans_{start_scan_id}_to_{stop_scan_id}.dat"
        )

    def _trigger_workflow_on_new_scan(self, scan: BlissScanType | None) -> dict | None:
        workflow = self._get_workflow()
        workflow_inputs = self._get_workflow_inputs(scan)

        future = submit(
            args=(workflow,),
            kwargs={"inputs": workflow_inputs, "outputs": [{"all": False}]},
        )

        clear_plot = self._get_parameter("_clear_plot")
        # Consolidate the workflow inputs into a single dictionary.
        workflow_inputs = {
            workflow_input["name"]: workflow_input["value"]
            for workflow_input in workflow_inputs
        }
        self.handler.handle_workflow_result(
            future.task_id, clear_plot=clear_plot, workflow_inputs=workflow_inputs
        )

    def __call__(
        self,
        start_scan_id: int,
        stop_scan_id: int,
        clear_plot: bool = True,
        save_txt: bool = False,
    ) -> None:
        if start_scan_id > stop_scan_id:
            logger.warning(
                f"Invalid scan indices; the start_scan_id ({start_scan_id}) "
                f"should be smaller than the stop_scan_id ({stop_scan_id})."
            )
            return

        # Temporarily set the scan ids for the workflow trigger.
        prev_start_scan_id = self._get_start_scan_id()
        prev_stop_scan_id = self._get_stop_scan_id()

        self._set_start_scan_id(start_scan_id)
        self._set_stop_scan_id(stop_scan_id)

        if clear_plot:
            self._set_parameter("_clear_plot", True)
        else:
            self._set_parameter("_clear_plot", False)

        try:
            self._trigger_workflow_on_new_scan(None)
        except Exception as e:
            logger.warning(e, exc_info=True)

        if save_txt:
            result = self.handler.get_result()
            if result is None:
                logger.warning("No aggregated data available to save.")
                return
            x = result["measurement"]["x"]
            signal = result["measurement"]["signal"]
            filename = self._get_processed_data_path()
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            np.savetxt(filename, np.column_stack((x, signal)), header="x\tsignal")
            print(f"Data saved to {filename}.")

        # Restore the previous scan ids.
        self._set_start_scan_id(prev_start_scan_id)
        self._set_stop_scan_id(prev_stop_scan_id)


class AggregatedScansHandler(WithFlintAccess):
    def __init__(self) -> None:
        super().__init__()
        self._workflow_inputs = None
        self._result = None

    def clear(self):
        plot = self._get_plot()
        plot.submit("clear")
        plot.xlabel = "X"
        plot.ylabel = "Y"
        plot.title = ""

    def handle_workflow_result(self, *args, **kwargs) -> None:
        """Handle workflow result in a background task to recover data."""
        gevent.spawn(self._handle_workflow_result, *args, **kwargs)

    def _handle_workflow_result(
        self, task_id, timeout: int = 60, clear_plot: bool = True, workflow_inputs=None
    ) -> None:
        logger.info(f"Recovering processed data for task {task_id}")
        try:
            result = get_future(task_id).get(timeout=timeout)
            if result:
                self._result = result
                # Merge workflow inputs back into the result for reference.
                self._workflow_inputs = workflow_inputs
                self.plot_data(clear_plot=clear_plot)
            else:
                self._result = None
        except Exception as e:
            logger.exception(e)
            self._result = None

    def _get_plot(self, plot_id="Aggregated Scans", plot_cls=AggregatedScansPlot):
        return super()._get_plot(plot_id, plot_cls)

    def _get_filename(self):
        if self._workflow_inputs is None:
            return ""
        filename = self._workflow_inputs.get("filename", "")
        return filename

    def _get_axes_labels(self) -> dict[str, str]:
        axes_labels = {"x": "X", "y": "Y"}

        if not self._workflow_inputs:
            return axes_labels

        data = self._workflow_inputs.get("data_mappings", {})
        if not data:
            return axes_labels

        if "x" in data:
            axes_labels["x"] = data["x"].split("/")[-1]

        if "signal" in data:
            axes_labels["y"] = data["signal"].split("/")[-1]
            if "monitor" in data:
                axes_labels["y"] += " / " + data["monitor"].split("/")[-1]

        return axes_labels

    def get_result(self) -> dict[str, Any] | None:
        return self._result

    def plot_data(self, clear_plot: bool = True):
        if not self._result or self._workflow_inputs is None:
            return

        plot = self._get_plot()
        if clear_plot:
            plot.submit("clear")

        scans = self._result["scans"]
        for scan_id in scans:
            color = "gray"
            legend = f"Scan {scan_id}"

            if scan_id == max(scans):
                color = "black"
                legend = legend + " (last)"

            plot.add_curve(
                scans[scan_id]["x"],
                scans[scan_id]["signal"],
                legend=legend,
                linestyle="-",
                color=color,
            )

        measurement = self._result["measurement"]
        plot.add_curve(
            measurement["x"],
            measurement["signal"],
            legend=self._workflow_inputs["aggregation_mode"].capitalize(),
            linestyle="-",
            color="red",
            linewidth=2,
        )

        filename = self._get_filename()
        plot.title = f"{filename}"

        axes_labels = self._get_axes_labels()
        plot.xlabel = axes_labels["x"]
        plot.ylabel = axes_labels["y"]



import pathlib

from ._assert import demo_assert


@demo_assert("Scan #{scan_number} in SPEC file {filename}")
def assert_spec_scan_exists(filename: pathlib.Path, scan_number: int) -> None:
    find_string = f"#S {scan_number}"
    with open(filename, "r") as f:
        for line in f:
            if find_string in line:
                break
        else:
            raise AssertionError(f"{find_string!r} not found in {str(filename)!r}")



import logging
import os
import uuid
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from ewoksjob.client import get_future
from ewoksjob.client import submit

from ..bliss_globals import current_session
from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..utils import directories

_logger = logging.getLogger(__name__)


class SinogramProcessor(
    BaseProcessor,
    parameters=[
        ParameterInfo("sleep_time", category="sinogram"),
        ParameterInfo("deltabeta", category="sinogram"),
        ParameterInfo("backends", category="sinogram"),
        ParameterInfo("cor_backend", category="sinogram"),
        # TODO: Param should be common?
        ParameterInfo("queue", category="workflows"),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("_enabled", True)
        defaults.setdefault("sleep_time", 0)
        defaults.setdefault("deltabeta", 200)
        defaults.setdefault("backends", "nabu,silx")
        defaults.setdefault("cor_backend", "bliss-cor8")

        # Set tomo wide defaults
        defaults.setdefault("trigger_at", "END")
        defaults.setdefault("queue", "tomo_sinogram")

        super().__init__(config=config, defaults=defaults)

    # TODO: Should move to common location
    def workflow_destination(self, scan: BlissScanType) -> str:
        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        root = self.scan_processed_directory(scan)
        stem = os.path.splitext(os.path.basename(filename))[0]
        basename = f"{stem}_{scan_nb:04d}.json"
        return os.path.join(root, basename)

    # TODO: Should move to common location
    def master_output_filename(self, scan: BlissScanType) -> str:
        """Filename which can be used to inspect the results after the processing."""
        filename = self.get_filename(scan)
        root = self.scan_processed_directory(scan)
        basename = os.path.basename(filename)
        return os.path.join(root, basename)

    # TODO: Should move to common location
    def get_filename(self, scan: BlissScanType) -> str:
        filename = scan.scan_info.get("filename")
        if filename:
            return filename
        return current_session.scan_saving.filename

    # TODO: Common
    def scan_processed_directory(self, scan: BlissScanType) -> str:
        return directories.get_dataset_processed_dir(self.get_filename(scan))

    def get_sinogram_inputs(
        self,
        sleep_time: Optional[float] = None,
        axisposition: Optional[float] = None,
        deltabeta: Optional[float] = None,
        backends: Optional[str] = None,
        cor_backend: Optional[str] = None,
    ):
        def arg_else_prop(arg, prop):
            return arg if arg is not None else prop

        params = {
            "sleep_time": arg_else_prop(sleep_time, self.sleep_time),
            "axisposition": axisposition,
            "deltabeta": arg_else_prop(deltabeta, self.deltabeta),
            "backends": arg_else_prop(backends, self.backends),
            "cor_backend": arg_else_prop(cor_backend, self.cor_backend),
        }

        inputs = [
            {
                "task_identifier": "SinogramReconstruction",
                "name": p,
                "value": v,
            }
            for p, v in params.items()
        ]

        return inputs

    def get_inputs(self, scan) -> List[dict]:
        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        inputs = [
            {
                "task_identifier": "SinogramReconstruction",
                "name": "filename",
                "value": filename,
            },
            {
                "task_identifier": "SinogramReconstruction",
                "name": "output_filename",
                "value": self.master_output_filename(scan),
            },
            {
                "task_identifier": "SinogramReconstruction",
                "name": "scan",
                "value": scan_nb,
            },
        ]
        inputs += self.get_sinogram_inputs()
        return inputs

    def get_reprocess_inputs(
        self,
        datacollectionid: int,
        filename: str,
        deltabeta: Optional[float],
        axisposition: Optional[float],
        overwrite: bool = False,
    ) -> List[dict]:

        process_root = directories.get_dataset_processed_dir(filename)
        if os.path.exists(process_root):
            job_uuid = str(uuid.uuid4())
            process_root = f"{process_root}_{job_uuid}"
            os.makedirs(process_root)

        basename = os.path.basename(filename)

        inputs = [
            {
                "task_identifier": "SinogramReconstruction",
                "name": "dataCollectionId",
                "value": datacollectionid,
            },
            {
                "task_identifier": "SinogramReconstruction",
                "name": "filename",
                "value": filename,
            },
            {
                "task_identifier": "SinogramReconstruction",
                "name": "output_filename",
                "value": os.path.join(process_root, basename),
            },
        ]
        inputs += self.get_sinogram_inputs(
            deltabeta=deltabeta, axisposition=axisposition
        )
        return inputs

    def get_submit_arguments(self, scan: BlissScanType) -> dict:
        return {
            "inputs": self.get_inputs(scan),
            "outputs": [{"all": False}],
        }

    def get_workflow(self):
        return {
            "graph": {
                "id": "tomo-sinogram-reconstruction",
                "label": "tomo sinogram reconstruction",
            },
            "nodes": [
                {
                    "id": "SinogramReconstruction",
                    "task_type": "class",
                    "task_identifier": "tomovis.ewoks.tomo_sinogram_reconstruction.TomoSinogramReconstruction",
                }
            ],
            "links": [],
        }

    def _trigger_workflow_on_new_scan(self, scan):
        _logger.debug("Trigger workflow on new scan")
        return self.on_new_scan_metadata(scan)

    def scan_requires_processing(self, scan: BlissScanType) -> bool:
        channels = scan.scan_info.get("channels", {})
        sinogram_chan = channels.get("sinogram") is not None
        rotation_chan = channels.get("rotation") is not None
        translation_chan = channels.get("translation") is not None
        sinogram = sinogram_chan and rotation_chan and translation_chan
        return sinogram

    def on_new_scan_metadata(self, scan) -> Optional[dict]:
        metadata, _ = self._on_new_scan(scan)
        return metadata

    def _on_new_scan(self, scan) -> Tuple[Optional[dict], Optional[Any]]:
        _logger.debug("_on_new_scan")

        if not self.scan_requires_processing(scan):
            return None, None

        workflow = self.get_workflow()
        kwargs = self.get_submit_arguments(scan)
        if scan.scan_info.get("save"):
            kwargs["convert_destination"] = self.workflow_destination(scan)

        _logger.debug("Execute workflow\n%s\nArgs:\n%s", workflow, kwargs)
        _logger.debug("Queue: %s", self.queue)

        future = submit(args=(workflow,), kwargs=kwargs, queue=self.queue)
        future = get_future(future.uuid)

        return None, future

    def reprocess(
        self,
        params: Dict[str, Any],
    ):
        _logger.debug("Reprocess: %s", params)

        datacollectionid = params.pop("datacollectionid")
        deltabeta = params.pop("deltabeta")
        axisposition = params.pop("axisposition")
        filename = params.pop("filename")
        overwrite = params.pop("overwrite", False)
        if len(params) != 0:
            _logger.error("Unexpected reprocess parameters: %s", params)

        workflow = self.get_workflow()
        kwargs = {
            "inputs": self.get_reprocess_inputs(
                axisposition=axisposition,
                deltabeta=deltabeta,
                filename=filename,
                datacollectionid=datacollectionid,
                overwrite=overwrite,
            ),
            "outputs": [{"all": False}],
        }

        future = submit(args=(workflow,), kwargs=kwargs, queue=self.queue)
        future = get_future(future.uuid)



from ...bm02.xrpd_processor import Bm02XrpdProcessor
from ...bm23.exafs_processor import Bm23ExafsProcessor
from ...exafs.processor import ExafsProcessor
from ...id11.xrpd_processor import Id11XrpdProcessor
from ...id12.converter import Id12Hdf5ToAsciiConverter
from ...id14.converter import Id14Hdf5ToSpecConverter
from ...id22.xrpd_processor import Id22XrpdProcessor
from ...id31.streamline_scanner import Id31StreamlineScanner
from ...id31.xrpd_processor import Id31XrpdProcessor
from ...id32.processor import Id32SpecGenProcessor
from ...streamline.scanner import StreamlineScanner
from ...wrappers.ewoks_macros import EwoksMacroHandler
from ...xrpd.processor import XrpdProcessor
from ..user_scripts.bm02 import bm02_xrpd_processor
from ..user_scripts.id11 import id11_xrpd_processor
from ..user_scripts.id22 import id22_xrpd_processor
from ..user_scripts.id22 import stscan_processor
from ..user_scripts.id31 import id31_xrpd_processor
from ..user_scripts.streamline import streamline_scanner
from .itest_ewoks_macros import EWOKS_MACROS
from .itest_exafs import EXAFS_PROCESSOR
from .itest_id12 import ID12_CONVERTER
from .itest_id14 import ID14_CONVERTER
from .itest_xrpd import XRPD_PROCESSED


def all_print():
    _print_objects(EXAFS_PROCESSOR)
    _print_objects(XRPD_PROCESSED)
    _print_objects(id11_xrpd_processor)
    _print_objects(id22_xrpd_processor)
    _print_objects(ID14_CONVERTER)
    _print_objects(ID12_CONVERTER)
    _print_objects(stscan_processor)
    _print_objects(streamline_scanner)
    _print_objects(bm02_xrpd_processor)
    _print_objects(id31_xrpd_processor)
    _print_objects(EWOKS_MACROS)

    _print_objects(XrpdProcessor())
    _print_objects(ExafsProcessor())
    _print_objects(StreamlineScanner())
    _print_objects(Bm23ExafsProcessor())
    _print_objects(Bm02XrpdProcessor())
    _print_objects(Id11XrpdProcessor())
    _print_objects(Id22XrpdProcessor())
    _print_objects(Id31XrpdProcessor())
    _print_objects(Id14Hdf5ToSpecConverter())
    _print_objects(Id12Hdf5ToAsciiConverter())
    _print_objects(Id31StreamlineScanner())
    _print_objects(Id32SpecGenProcessor())
    _print_objects(EwoksMacroHandler())


def _print_objects(obj):
    print()
    print("===================")
    print(obj._parameters.name)
    print(obj.__info__())



from typing import Any
from typing import Dict
from typing import Optional

from ...wrappers.ewoks_macros import EwoksMacroHandler


class DemoEwoksMacroHandler(EwoksMacroHandler):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        if defaults is None:
            defaults = {}
        defaults.setdefault("queue", "celery")
        super().__init__(config, defaults)



"""
.. code-block:: python

    DEMO_SESSION [1]: from blissoda.demo.streamline_scanner import streamline_scanner,sample_changer
    DEMO_SESSION [2]: streamline_scanner.eject()
    DEMO_SESSION [3]: streamline_scanner.load()
    DEMO_SESSION [4]: streamline_scanner.calib(1, sample_index=0)
    DEMO_SESSION [5]: streamline_scanner.run(0.1)
"""

import os
import re
import shutil
from contextlib import contextmanager
from numbers import Number
from typing import Any
from typing import Dict
from typing import Generator
from typing import NamedTuple
from typing import Optional
from typing import Tuple

import numpy
from esrf_ontologies import technique
from ewoksjob.client import submit
from ewoksutils.task_utils import task_inputs

from ..automation import BlissAutomationObject
from ..bliss_globals import current_session
from ..bliss_globals import setup_globals
from ..persistent.parameters import ParameterInfo
from ..resources import resource_filename
from ..utils import directories
from ..utils import validators
from ..utils.icat import adapt_legacy_metadata

try:
    from bliss.common.logtools import elog_print
except ImportError:
    elog_print = print


class ScanInfo(NamedTuple):
    filename: str
    scan_nb: int

    @property
    def url(self):
        return f"{self.filename}::/{self.scan_nb}.1"


class StreamlineScanner(
    BlissAutomationObject,
    parameters=[
        ParameterInfo("workflow", category="data processing"),
        ParameterInfo("queue", category="data processing", validator=str),
        ParameterInfo("sample_changer_name", category="names"),
        ParameterInfo("detector_name", category="names", validator=str),
        ParameterInfo("energy_name", category="names"),
        ParameterInfo("calibration_scans", category="calibration"),
        ParameterInfo("calibration_motor", category="calibration"),
        ParameterInfo("image_slice", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
        ParameterInfo("pyfai_config", category="PyFai", validator=validators.is_file),
        ParameterInfo("calibrant", category="calibration"),
        ParameterInfo("trigger_workflows", category="data processing", validator=bool),
        ParameterInfo("vibration_speed_during_scan", category="sample changer"),
        ParameterInfo("verify_qrcode", category="robust vs. speed", validator=bool),
        ParameterInfo(
            "skip_when_no_qr_code", category="robust vs. speed", validator=bool
        ),
        ParameterInfo("autotune_qrreader_per", category="robust vs. speed"),
        ParameterInfo("dryrun", category="testing", validator=bool),
        ParameterInfo("calib_ring_detector_name", category="calibration"),
    ],
):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("queue", "online")
        defaults.setdefault("image_slice", 0)
        defaults.setdefault("trigger_workflows", True)
        defaults.setdefault("vibration_speed_during_scan", 40)
        defaults.setdefault("dryrun", False)
        defaults.setdefault("verify_qrcode", False)
        defaults.setdefault("skip_when_no_qr_code", False)
        defaults.setdefault("autotune_qrreader_per", "baguette")

        super().__init__(config=config, defaults=defaults)

        self._technique_metadata = technique.get_technique_metadata("XRPD")

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        try:
            sample_changer = self.sample_changer
        except (AttributeError, KeyError):
            return categories
        categories["status"] = {
            "# sample holders left": sample_changer.number_of_remaining_baguettes,
            "selected sample": int(sample_changer.translation.position),
            "vibration speed (%)": sample_changer.vibration_speed,
            "automatic calibration": self.workflow_has_calib,
            "calibration": "OFF",
            "flat-field": "OFF",
        }
        if self.workflow_has_calib:
            categories["status"]["calibration"] = self._get_calibration().get("image")
        if self.workflow_has_flatfield:
            categories["status"]["flat-field"] = "ON"
        return categories

    @property
    def autotune_qrreader_per(self) -> Optional[str]:
        return self._get_parameter("autotune_qrreader_per")

    @autotune_qrreader_per.setter
    def autotune_qrreader_per(self, value: Optional[str]):
        if value not in (None, "sample", "baguette"):
            raise ValueError("Allowed values are 'sample', 'baguette' or None")
        self._set_parameter("autotune_qrreader_per", value)

    def measure_sample(self, *args, has_qrcode: bool = True, **kwargs):
        if self.dryrun:
            print("Dry-run: skip measurement")
        else:
            return setup_globals.sct(*args, **kwargs)

    @property
    def sample_changer(self):
        self._raise_when_missing("sample_changer_name")
        return current_session.env_dict[self.sample_changer_name]

    def eject(self):
        self.sample_changer.translation.on()
        self.sample_changer.eject_old_baguette()
        print(
            "\n\nNumber of remaining sample holders:",
            self.sample_changer.number_of_remaining_baguettes,
        )

    def load(self):
        self.sample_changer.translation.on()
        self.sample_changer.load_baguette_with_homing()
        if self.autotune_qrreader_per == "baguette":
            self.tune_qrreader_for_baguette()

    def run(
        self,
        *scan_args,
        nholders: Optional[int] = None,
        use_qr_code: bool = True,
        current_holder: bool = False,
        sample_indices: Optional[Tuple[int]] = None,
        **scan_kwargs,
    ):
        if self.workflow_has_calib and not self._get_calibration():
            raise RuntimeError("measure a calibration standard first")
        if self.trigger_workflows:
            if not self.workflow:
                raise RuntimeError("initialize a workflow first")
            if self.pyfai_config and not os.path.exists(self.pyfai_config):
                raise RuntimeError("the pyFAI configuration file no longer exists")

        with self.run_context():
            if current_holder:
                self.load()
                self._run_holder(
                    scan_args,
                    scan_kwargs,
                    use_qr_code=use_qr_code,
                    sample_indices=sample_indices,
                )
            elif nholders is None:
                while self.sample_changer.has_remaining_baguettes():
                    self.eject()
                    self.load()
                    self._run_holder(
                        scan_args,
                        scan_kwargs,
                        use_qr_code=use_qr_code,
                        sample_indices=sample_indices,
                    )
                self.eject()
            else:
                for _ in range(nholders):
                    self.eject()
                    self.load()
                    self._run_holder(
                        scan_args,
                        scan_kwargs,
                        use_qr_code=use_qr_code,
                        sample_indices=sample_indices,
                    )
                self.eject()

    def _run_holder(
        self,
        scan_args: tuple,
        scan_kwargs: dict,
        use_qr_code: bool = True,
        sample_indices: Optional[Tuple[int]] = None,
    ):
        if use_qr_code:
            itfunc = self.sample_changer.iterate_samples
        else:
            itfunc = self.sample_changer.iterate_samples_without_qr

        print("")
        print("========== RUN HOLDER ==========")
        for qrcode in itfunc(
            sample_indices=sample_indices,
            autoTuningAllowed=self.autotune_qrreader_per == "sample",
        ):
            print()
            self._process_sample(qrcode, scan_args, scan_kwargs)
        print("================================")

    def calib(
        self,
        *scan_args,
        sample_index: Optional[int] = None,
        use_qr_code: bool = True,
        **scan_kwargs,
    ):
        if sample_index is None:
            raise ValueError("argument 'sample_index' not provided")
        with self.run_context():
            qrcode = self._select_sample(sample_index, use_qr_code=use_qr_code)
            self._process_sample(qrcode, scan_args, scan_kwargs, is_calibration=True)

    def select_sample(self, sample_index: int, use_qr_code: bool = True) -> str:
        self.sample_changer.translation.on()
        return self._select_sample(sample_index, use_qr_code=use_qr_code)

    def _select_sample(self, sample_index: int, use_qr_code: bool = True) -> str:
        if use_qr_code:
            select_sample = self.sample_changer.select_sample
        else:
            select_sample = self.sample_changer.select_sample_without_qr
        return select_sample(
            sample_index, autoTuningAllowed=self.autotune_qrreader_per == "sample"
        )

    def init_workflow(self, with_autocalibration: bool = False):
        if with_autocalibration:
            basename = "streamline_with_calib"
        else:
            basename = "streamline_without_calib"
        filename = f"{basename}.json"

        dirname = self._get_workflows_dir(current_session.scan_saving.filename)
        destination = os.path.join(dirname, filename)
        if not os.path.exists(destination):
            os.makedirs(dirname, exist_ok=True)
            source = resource_filename("streamline", filename)
            shutil.copyfile(source, destination)

        self.workflow = destination
        print(f"Active data processing workflow: {destination}")

    @property
    def workflow_has_calib(self):
        return self.workflow and "with_calib" in self.workflow

    @property
    def workflow_has_flatfield(self):
        return self.workflow and "with_flat" in self.workflow

    def _set_calibration_scan(self, scan_info: ScanInfo):
        if not scan_info.filename:
            print("Cannot use as calibration because no data was collected")
            return
        info = {
            "image": self._get_image_url(scan_info),
            "gallery_directory": self._get_gallery_directory(scan_info.filename),
        }
        position = self._get_calibration_position()
        if self.calibration_scans is None:
            self.calibration_scans = dict()
        self.calibration_scans[position] = info

    def _trigger_processing(self, scan_info: ScanInfo, processed_metadata: dict):
        if not scan_info.filename:
            print("Cannot trigger workflow because no data was collected")
            return
        args, kwargs = self._job_arguments(scan_info, processed_metadata)
        submit(args=args, kwargs=kwargs, queue=self.queue)

    @contextmanager
    def run_context(self):
        self.sample_changer.translation.on()
        self.sample_changer.vibration_speed = self.vibration_speed_during_scan
        elog_print("Start streamline run")
        try:
            yield
        finally:
            self.sample_changer.vibration_speed = 0
            elog_print("End streamline run")

    def _process_sample(
        self,
        qrcode: str,
        scan_args: tuple,
        scan_kwargs: dict,
        is_calibration: bool = False,
    ):
        try:
            with self._verify_qrcode(qrcode) as qrcode:
                has_qrcode = qrcode != self._qrcode_error

                if not has_qrcode and self.skip_when_no_qr_code:
                    elog_print("SKIP SAMPLE (NO QR-CODE)")
                    return

                self._newsample(qrcode)

                self._set_scan_metadata(scan_args, scan_kwargs)
                self._set_raw_dataset_metadata(scan_args, scan_kwargs)

                scan = self.measure_sample(
                    *scan_args, has_qrcode=has_qrcode, **scan_kwargs
                )

                scan_info = self._get_scan_info(scan)
                if is_calibration:
                    self._set_calibration_scan(scan_info)

                if self.trigger_workflows:
                    processed_metadata = self._get_processed_dataset_metadata(scan_args)
                    if self.dryrun:
                        print("Dry-run: skip workflow triggering")
                    else:
                        self._trigger_processing(scan_info, processed_metadata)

                if is_calibration:
                    elog_print(
                        f"Streamline calibrant {self.calibrant}: {current_session.scan_saving.filename}"
                    )
        finally:
            setup_globals.enddataset()

    def qr_read(self) -> str:
        return self.sample_changer.qr_read(
            autoTuningAllowed=self.autotune_qrreader_per == "sample"
        )

    @property
    def _qrcode_error(self) -> str:
        return self.sample_changer.qrreader.QRCODE_NOT_READABLE

    def tune_qrreader(self, force=False) -> str:
        self.sample_changer.tune_qrreader(force=force)

    def tune_qrreader_for_baguette(self) -> None:
        for _ in self.sample_changer.iterate_samples_without_qr():
            qrcode = self.tune_qrreader()
            if qrcode != self._qrcode_error:
                break
        else:
            print("QR-reader tuning failed when loading the baguette")

    @contextmanager
    def _verify_qrcode(self, qrcode: str) -> Generator[str, None, None]:
        """Check the QR-code before (yields the new code when it changed)
        and check the QR-code after (raises and exception when it changed)"""
        if not self.verify_qrcode:
            yield qrcode
            return

        qrcode_now = self.qr_read()
        if qrcode_now != qrcode and qrcode_now != self._qrcode_error:
            print(
                f"Reading the QR-code twice gives first '{qrcode}' and then '{qrcode_now}'"
            )
            qrcode = qrcode_now
        try:
            yield qrcode
        finally:
            qrcode_now = self.qr_read()
            if qrcode_now != qrcode and qrcode_now != self._qrcode_error:
                if qrcode == self._qrcode_error:
                    msg = f"The sample name of dataset '{current_session.scan_saving.filename}' is '{qrcode_now}' (read at the end of the dataset)"
                    elog_print(msg)
                else:
                    msg = f"The sample name of dataset '{current_session.scan_saving.filename}' might be '{qrcode_now}' (read at the end of the dataset)"
                    elog_print(msg)
                    raise RuntimeError(msg)

    def _newsample(self, qrcode: str) -> None:
        setup_globals.newsample(qrcode)
        setup_globals.newdataset()

    def _set_raw_dataset_metadata(self, scan_args: tuple, scan_kwargs: dict) -> None:
        for k, v in self._get_raw_dataset_metadata(scan_args).items():
            current_session.scan_saving.dataset[k] = v

    def _set_scan_metadata(self, scan_args: tuple, scan_kwargs: dict) -> None:
        scan_info = self._get_scan_metadata()
        if scan_info:
            scan_kwargs["scan_info"] = scan_info

    def _get_scan_metadata(self) -> dict:
        return self._technique_metadata.get_scan_info()

    def _get_raw_dataset_metadata(self, scan_args: tuple) -> dict:
        metadata = self._technique_metadata.get_dataset_metadata()
        if self.energy_name:
            metadata["HTXRPD_energy"] = getattr(
                setup_globals, self.energy_name
            ).position
        if scan_args and isinstance(scan_args[0], Number):
            metadata["HTXRPD_exposureTime"] = scan_args[0]
        speed = self.vibration_speed_during_scan
        if speed is not None:
            metadata["HTXRPD_sampleVibration"] = speed
        position = self._get_calibration_position()
        if position is not None:
            metadata["HTXRPD_distance"] = position
        adapt_legacy_metadata(metadata)
        return metadata

    def _get_processed_dataset_metadata(self, scan_args: tuple) -> dict:
        metadata = self._technique_metadata.get_dataset_metadata()
        metadata["Sample_name"] = current_session.scan_saving.dataset["Sample_name"]
        return metadata

    def _get_calibration_position(self) -> Optional[Number]:
        if self.calibration_motor:
            return getattr(setup_globals, self.calibration_motor).position

    def _get_calibration(self) -> dict:
        calibration_scans = self.calibration_scans
        if not calibration_scans:
            return dict()
        position = self._get_calibration_position()
        if position is not None:
            positions = [p for p in calibration_scans if p is not None]
            if positions:
                idx = (numpy.abs(numpy.array(positions) - position)).argmin()
                position = positions[idx]
        return calibration_scans.get(position, dict())

    def _get_scan_info(self, scan) -> ScanInfo:
        if scan is None:
            return ScanInfo(filename="", scan_nb=0)
        if isinstance(scan, ScanInfo):
            return scan
        filename = scan.scan_info.get("filename")
        scan_nb = scan.scan_info.get("scan_nb")
        return ScanInfo(filename=filename, scan_nb=scan_nb)

    def _get_image_url(self, scan_info: ScanInfo) -> str:
        url = f"silx://{scan_info.filename}?path=/{scan_info.scan_nb}.1/measurement/{self.detector_name}"
        image_slice = self.image_slice
        if image_slice is not None:
            image_slice = str(image_slice)
            image_slice = re.sub(r"[\s\(\)]+", "", image_slice)
            url = f"{url}&slice={image_slice}"
        return url

    def _get_output_dir(self, dataset_filename: str) -> str:
        return os.path.join(
            directories.get_processed_dir(dataset_filename), "streamline"
        )

    def _get_transient_dirname(self, dataset_filename: str) -> str:
        return directories.get_nobackup_dir(dataset_filename)

    def _get_workflows_dir(self, dataset_filename: str) -> str:
        return directories.get_workflows_dir(dataset_filename)

    def _get_output_dirname(self, dataset_filename: str) -> str:
        filename = os.path.basename(dataset_filename)
        return os.path.join(
            self._get_output_dir(dataset_filename), os.path.splitext(filename)[0]
        )

    def _get_gallery_dirname(self, dataset_filename: str) -> str:
        return os.path.join(self._get_output_dirname(dataset_filename), "gallery")

    def _get_hdf5_output_filename(self, dataset_filename: str) -> str:
        return os.path.join(
            self._get_output_dirname(dataset_filename),
            os.path.basename(dataset_filename),
        )

    def _get_ascii_output_filename(self, dataset_filename: str, unit: str) -> str:
        basename, _ = os.path.splitext(os.path.basename(dataset_filename))
        return os.path.join(
            self._get_output_dirname(dataset_filename), f"{basename}_{unit}.xye"
        )

    def _get_gallery_directory(self, dataset_filename: str) -> str:
        return self._get_gallery_dirname(dataset_filename)

    def _get_workflow_save_filename(self, dataset_filename: str) -> str:
        basename = os.path.basename(self.workflow)
        return os.path.join(
            self._get_output_dirname(dataset_filename),
            basename,
        )

    def _get_workflow_upload_parameters(
        self, dataset_filename: str, processed_metadata: dict
    ) -> Optional[dict]:
        raw = os.path.dirname(dataset_filename)
        dataset = "integrate"
        scan_saving = current_session.scan_saving
        proposal = scan_saving.proposal_name
        beamline = scan_saving.beamline
        path = self._get_output_dirname(dataset_filename)
        return {
            "beamline": beamline,
            "proposal": proposal,
            "dataset": dataset,
            "path": path,
            "raw": [raw],
            "metadata": processed_metadata,
        }

    def _job_arguments(self, scan_info: ScanInfo, processed_metadata: dict):
        """Arguments for the workflow execution"""
        self._raise_when_missing("workflow")
        if self.workflow_has_calib:
            self._raise_when_missing("calibration_scans", "calibrant")

        inputs = list()

        integrate_image_url = self._get_image_url(scan_info)

        # Configuration
        if self.integration_options:
            inputs.append(
                {
                    "task_identifier": "PyFaiConfig",
                    "name": "integration_options",
                    "value": self.integration_options.to_dict(),
                }
            )
        if self.pyfai_config:
            inputs.append(
                {
                    "task_identifier": "PyFaiConfig",
                    "name": "filename",
                    "value": self.pyfai_config,
                }
            )
        if self.calibrant:
            inputs.append(
                {
                    "task_identifier": "PyFaiConfig",
                    "name": "calibrant",
                    "value": self.calibrant,
                }
            )

        # Calibration
        if self.workflow_has_calib:
            calibration = self._get_calibration()
            if not calibration:
                raise RuntimeError("no valid calibration found")

            inputs += task_inputs(
                task_identifier="CalibrateSingle",
                inputs={
                    "image": calibration["image"],
                    "fixed": ["energy"],
                    "robust": False,
                    "ring_detector": self.calib_ring_detector_name,
                },
            )
            inputs += task_inputs(
                task_identifier="DiagnoseCalibrateSingleResults",
                inputs={
                    "image": calibration["image"],
                    "filename": os.path.join(
                        calibration["gallery_directory"], "ring_detection.png"
                    ),
                },
            )

            if calibration["image"] == integrate_image_url and self.calibrant:
                inputs += [
                    {
                        "task_identifier": "DiagnoseIntegrate1D",
                        "name": "calibrant",
                        "value": self.calibrant,
                    },
                ]

        # Integration
        inputs += task_inputs(
            task_identifier="Integrate1D",
            inputs={
                "image": integrate_image_url,
            },
        )
        inputs += task_inputs(
            task_identifier="SaveNexusPattern1D",
            inputs={
                "url": self._get_hdf5_output_filename(scan_info.filename),
                "bliss_scan_url": scan_info.url,
            },
        )

        # Different outputs depending on the unit
        detector_name = self.detector_name
        for unit in ("q", "2th"):
            inputs += task_inputs(
                task_identifier="SaveNexusPattern1D",
                label=f"save_{unit}_hdf5",
                inputs={
                    "nxprocess_name": f"{detector_name}_integrate_{unit}",
                    "nxmeasurement_name": f"{detector_name}_integrated_{unit}",
                    "metadata": {
                        f"{detector_name}_integrate_{unit}": {
                            "configuration": {"workflow": self.workflow}
                        }
                    },
                },
            )
            inputs.append(
                {
                    "task_identifier": "SaveAsciiPattern1D",
                    "label": f"save_{unit}_ascii",
                    "name": "filename",
                    "value": self._get_ascii_output_filename(scan_info.filename, unit),
                }
            )

        inputs += [
            {
                "task_identifier": "DiagnoseIntegrate1D",
                "name": "filename",
                "value": os.path.join(
                    self._get_gallery_directory(scan_info.filename),
                    "integrate.png",
                ),
            },
        ]

        # Job arguments
        args = (self.workflow,)
        convert_destination = self._get_workflow_save_filename(scan_info.filename)
        upload_parameters = self._get_workflow_upload_parameters(
            scan_info.filename, processed_metadata
        )
        if self.workflow_has_calib:
            varinfo = {
                "root_uri": self._get_transient_dirname(scan_info.filename),
                "scheme": "nexus",
            }
        else:
            varinfo = None
        kwargs = {
            "engine": None,
            "inputs": inputs,
            "varinfo": varinfo,
            "convert_destination": convert_destination,
            "upload_parameters": upload_parameters,
            "save_options": {"indent": 2},
        }
        return args, kwargs



import logging
from contextlib import contextmanager
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Type

from ..bliss_globals import current_session
from ..import_utils import is_available
from ..import_utils import unavailable_class
from ..import_utils import unavailable_function
from . import BasePlot
from . import FlintClient

try:
    from gevent.lock import RLock
except ImportError as ex:
    RLock = unavailable_class(ex)

try:
    from bliss.common.plot import get_flint as _native_get_flint
except ImportError as ex:
    _native_get_flint = unavailable_function(ex)


logger = logging.getLogger(__name__)


class WithFlintAccess:
    _HAS_BLISS = is_available(current_session)

    def __init__(self) -> None:
        self.__plots: Dict[str, BasePlot] = dict()
        self.__flint_client_id = None
        self.__plots_lock = None
        self.__get_plot_stack = 0

    @property
    def _plots_lock(self) -> RLock:
        if self.__plots_lock is None:
            self.__plots_lock = RLock()
        return self.__plots_lock

    def _get_plot(self, plot_id: str, plot_cls: Type[BasePlot]) -> BasePlot:
        """Launches Flint and creates the plot when either is missing."""
        with self._plots_lock:
            with self._access_flint_context() as flint_client:
                plot = self.__plots.get(plot_id)
                if plot is None:
                    plot = flint_client.get_plot(plot_cls, unique_name=plot_id)
                    logger.info("Created Flint plot %r", plot_id)
                    self.__plots[plot_id] = plot
                return plot

    @contextmanager
    def _access_flint_context(
        self, reset: bool = False
    ) -> Generator[FlintClient, None, None]:
        with self._plots_lock:
            self.__get_plot_stack += 1
            try:
                flint_client = _get_flint(reset=reset)
                flint_client_id = id(flint_client), flint_client.pid
                if flint_client_id != self.__flint_client_id:
                    self.__flint_client_id = flint_client_id
                    if self.__get_plot_stack > 1:
                        if self.__plots:
                            logger.warning(
                                "Cannot resetting Flint plots after restart (recursive call)"
                            )
                    else:
                        self._on_flint_restart(flint_client)
                yield flint_client
            finally:
                self.__get_plot_stack -= 1

    def reset_flint(self) -> None:
        with self._access_flint_context(reset=True) as _:
            pass

    def _on_flint_restart(self, flint_client: FlintClient) -> None:
        """Called whenever a new Flint client is instantiated."""
        with self._plots_lock:
            plots = {}
            for plot_id in self.__plots:
                plot_instance = self.__plots[plot_id]
                plot_cls = type(plot_instance)
                plots[plot_id] = flint_client.get_plot(plot_cls, unique_name=plot_id)
            self.__plots = plots


_flint_lock: Optional[RLock] = None
_flint_client: Optional[FlintClient] = None
_flint_pid: Optional[int] = None


def _get_flint(reset: bool = False) -> FlintClient:
    """Create the Flint client when needed:

    - not created yet
    - different process id
    - not available
    """
    global _flint_client, _flint_pid, _flint_lock
    if _flint_lock is None:
        _flint_lock = RLock()

    with _flint_lock:
        if reset:
            _flint_client = None
            _flint_pid = None

        try:
            new_client = (
                _flint_client is None
                or _flint_pid != _flint_client.pid
                or not _flint_client.is_available()
            )
        except FileNotFoundError:
            new_client = True

        if new_client:
            _flint_client = _native_get_flint()
            _flint_pid = _flint_client.pid
        return _flint_client



"""EXAFS plotting in Flint."""

import logging
import os
from collections import OrderedDict
from typing import Dict
from typing import List
from typing import Sequence
from typing import Tuple

from ewoksjob.client import submit

from ..flint import FlintClient
from ..flint.access import WithFlintAccess
from ..utils.directories import get_dataset_processed_dir
from .plots import ExafsPlot
from .types import ExafsPlotWorkflowParameters
from .types import ExafsSplitWorkflowParameters
from .types import ScanInfo
from .types import SubScanInfo
from .types import XasPlotData
from .types import XasSubscanData

logger = logging.getLogger(__name__)


class ExafsPlotter(WithFlintAccess):
    """Manage EXAFS plots in Flint based on Ewoks workflow results."""

    def __init__(self, number_of_scans: int = 0) -> None:
        super().__init__()
        self._number_of_scans = number_of_scans

        # Fixed parameters
        self._plot_names = {
            "flatten_mu": "mu",
            "chi_weighted_k": "chi",
            "ft_mag": "ft",
            "noise_savgol": "noise",
        }

        # Runtime data
        self._scan_infos: Dict[str, ScanInfo] = OrderedDict()
        self._color_index = 0

    @property
    def number_of_scans(self):
        """Maximum number of scans to be plotted"""
        return self._number_of_scans

    @number_of_scans.setter
    def number_of_scans(self, value):
        self._number_of_scans = max(value, 0)
        self.sync_plots()
        self.refresh()

    @property
    def scan_ids(self) -> List[str]:
        return list(self._scan_infos)

    def clear(self) -> None:
        """Remove all scan curves in all plots"""
        self._get_plot().clear()
        for scan_info in self._scan_infos.values():
            for subscan_info in scan_info.subscans:
                subscan_info.enabled = False

    def refresh(self) -> None:
        """Refresh all plots with the current processed data"""
        for scan_id, scan_info in self._scan_infos.items():
            for subscan_info in scan_info.subscans:
                subscan_info.updated = True
            self._update_scan_plot(scan_id)

    def create_scan_id(self, filename: str, scan_number: int) -> str:
        return f"{filename}|{scan_number}"

    def ensure_scan_infos(
        self, filename: str, scan_number: int
    ) -> Tuple[str, ScanInfo]:
        scan_id = self.create_scan_id(filename, scan_number)
        scan_info = self._scan_infos.get(scan_id)
        if scan_info:
            return scan_id, scan_info
        scan_info = ScanInfo(
            filename=filename, scan_number=scan_number, subscans=[], xas_results=[]
        )
        self._scan_infos[scan_id] = scan_info
        return scan_id, scan_info

    def _get_total_number_of_subscans(self) -> int:
        n_subscans = 0
        for scan_info in self._scan_infos.values():
            n_subscans += len(scan_info.subscans)

        return n_subscans

    def sync_plots(self):
        """Synchronize workflow results in cache with plots in Flint."""
        n_subscans = self._get_total_number_of_subscans()

        max_subscans = self.number_of_scans
        min_enabled_subscan_idx = max(n_subscans - max_subscans, 0)
        subscan_index = 0
        for scan_info in self._scan_infos.values():
            for subscan_info, xasplotdata in zip(
                scan_info.subscans, scan_info.xas_results
            ):
                enabled = subscan_index >= min_enabled_subscan_idx
                needs_refresh = subscan_info.updated or subscan_info.enabled != enabled
                if needs_refresh:
                    if enabled:
                        self._get_plot().update_scan(
                            subscan_info.legend, xasplotdata, subscan_info.color
                        )
                    else:
                        self._get_plot().remove_scan(subscan_info.legend)
                subscan_info.enabled = enabled
                subscan_info.updated = False
                subscan_index += 1

    def remove_scan(self, legend: str) -> None:
        """Disable subscan and remove from Flint."""
        removed = False
        for scan_info in self._scan_infos.values():
            for subscan_info in scan_info.subscans:
                if subscan_info.legend == legend:
                    if subscan_info.enabled:
                        subscan_info.enabled = False
                        self._get_plot().remove_scan(subscan_info.legend)
                        removed = True
                    break
        if removed:
            self.purge_scan_infos()

    def purge_scan_infos(self, keep_scan_ids: Sequence[str] = tuple()) -> None:
        """Remove cache from scans that have no enabled sub-scans or not processing results."""
        for scan_id, scan_info in list(self._scan_infos.items()):
            if scan_id in keep_scan_ids:
                continue

            # Disable all subscans when the scan has no data
            if not scan_info.xas_results:
                for subscan_info in scan_info.subscans:
                    if subscan_info.enabled:
                        subscan_info.enabled = False
                        self._get_plot().remove_scan(subscan_info.legend)

            # Delete the scan when it has no enabled subscans
            has_enabled_subscan = any(
                subscan_info.enabled for subscan_info in scan_info.subscans
            )
            if not has_enabled_subscan:
                del self._scan_infos[scan_id]

    def _update_scan_plot(self, scan_id: str) -> None:
        """Update all scan curves in all plots"""
        scan_info = self._scan_infos.get(scan_id)
        if scan_info is not None:
            self.sync_results(scan_info)
            self.sync_plots()

    def _on_flint_restart(self, flint_client: FlintClient) -> None:
        super()._on_flint_restart(flint_client)
        for scan_info in self._scan_infos.values():
            self.sync_results(scan_info)
        self.sync_plots()

    def sync_results(self, scan_info: ScanInfo) -> None:
        """Update workflow results in cache with processing results.
        Blocks when the scan has a pending future.
        """
        if scan_info.plot_future is None:
            return

        try:
            workflow_results = scan_info.plot_future.result()
        except Exception:
            # Workflow failed
            return
        finally:
            scan_info.plot_future = None

        xas_results = [
            XasSubscanData(
                **{
                    self._plot_names[plot_name]: XasPlotData(**plot_data)
                    for plot_name, plot_data in data.items()
                }
            )
            for data in workflow_results["plot_data"]
        ]

        if not xas_results:
            # Workflow succeeded but did not return any results
            return

        for idx, xasplotdata in enumerate(
            xas_results, scan_info.min_subscan_index_to_process
        ):
            if idx < len(scan_info.xas_results):
                scan_info.xas_results[idx] = xasplotdata
                scan_info.subscans[idx].updated = True
            else:
                basename = os.path.basename(os.path.dirname(scan_info.filename))
                legend = f"{basename}: {scan_info.scan_number}.{idx + 1}"
                color = self._new_color()
                subscan_info = SubScanInfo(legend=legend, color=color, enabled=False)
                scan_info.subscans.append(subscan_info)
                scan_info.xas_results.append(xasplotdata)

    def _submit_plot_workflow(
        self,
        scan_id: str,
        parameters: ExafsPlotWorkflowParameters,
        reprocess: bool = False,
    ) -> None:
        """Submit the data processing for a scan"""
        scan_info = self._scan_infos.get(scan_id, None)
        if scan_info is None:
            return

        scan_info.reprocess_all = reprocess
        input_information = {
            "channel_url": f"{scan_info.scan_url}/measurement/{parameters.energy_name}",
            "spectra_url": f"{scan_info.scan_url}/measurement/{parameters.mu_name}",
            "energy_unit": parameters.energy_unit,
        }

        if scan_info.multi_xas_scan:
            input_information["is_concatenated"] = True
            input_information["trim_concatenated_n_points"] = parameters.trim_n_points
            input_information["skip_concatenated_n_spectra"] = (
                scan_info.min_subscan_index_to_process
            )
            input_information["concatenated_spectra_section_size"] = (
                scan_info.multi_xas_subscan_size
            )

        plot_inputs = [
            {
                "task_type": "ReadXasObject",
                "name": "input_information",
                "value": input_information,
            },
            {
                "task_type": "PlotSpectrumData",
                "name": "plot_names",
                "value": list(self._plot_names),
            },
        ]
        scan_info.plot_future = submit(
            args=(parameters.workflow,), kwargs={"inputs": plot_inputs}
        )

    def _submit_split_workflow(
        self,
        scan_id: str,
        parameters: ExafsSplitWorkflowParameters,
        reprocess: bool = False,
    ) -> None:
        """Submit the data processing for a scan"""
        scan_info = self._scan_infos.get(scan_id, None)
        if scan_info is None or not scan_info.multi_xas_scan:
            return

        out_dirname = get_dataset_processed_dir(scan_info.filename)
        h5_basename = os.path.basename(scan_info.filename)
        h5_stem, _ = os.path.splitext(h5_basename)
        out_filename = os.path.join(out_dirname, h5_basename)
        convert_destination = os.path.join(
            out_dirname, "workflows", f"{h5_stem}_scan{scan_info.scan_number:04d}.json"
        )

        if out_filename.startswith("/data/scisoft/ewoks"):
            out_filename = out_filename.replace("/data/scisoft/ewoks", "/tmp_14_days")

        split_inputs = [
            {
                "task_type": "SplitBlissScan",
                "name": "filename",
                "value": scan_info.filename,
            },
            {
                "task_type": "SplitBlissScan",
                "name": "scan_number",
                "value": scan_info.scan_number,
            },
            {
                "task_type": "SplitBlissScan",
                "name": "monotonic_channel",
                "value": parameters.monotonic_channel,
            },
            {
                "task_type": "SplitBlissScan",
                "name": "subscan_size",
                "value": scan_info.multi_xas_subscan_size,
            },
            {
                "task_type": "SplitBlissScan",
                "name": "trim_n_points",
                "value": parameters.trim_n_points,
            },
            {
                "task_type": "SplitBlissScan",
                "name": "wait_finished",
                "value": parameters.scan_complete,
            },
            {
                "task_type": "SplitBlissScan",
                "name": "out_filename",
                "value": out_filename,
            },
            {
                "task_type": "SplitBlissScan",
                "name": "counter_group",
                "value": "measurement",
            },
        ]

        scan_info.split_future = submit(
            args=(parameters.workflow,),
            kwargs={"inputs": split_inputs, "convert_destination": convert_destination},
        )

    def execute_and_plot(
        self,
        scan_id: str,
        plot_parameters: ExafsPlotWorkflowParameters,
        split_parameters: ExafsSplitWorkflowParameters,
        reprocess: bool = False,
    ) -> None:
        self._submit_plot_workflow(scan_id, plot_parameters, reprocess=reprocess)
        self._submit_split_workflow(scan_id, split_parameters, reprocess=reprocess)
        self._update_scan_plot(scan_id)

    def _get_plot(self) -> ExafsPlot:
        """Launches Flint and creates the plot when either is missing"""
        return super()._get_plot("EXAFS", ExafsPlot)

    @property
    def _n_colors(self) -> int:
        return min(len(self._COLOR_PALETTE), self.number_of_scans + 1)

    def _new_color(self) -> Tuple[int, int, int]:
        scan_color = self._COLOR_PALETTE[min(self._color_index, self._n_colors - 1)]
        self._color_index = (self._color_index + 1) % self._n_colors
        return scan_color

    _COLOR_PALETTE = [
        (87, 81, 212),
        (235, 171, 33),
        (176, 69, 0),
        (0, 197, 248),
        (207, 97, 230),
        (0, 166, 107),
        (184, 0, 87),
        (0, 138, 248),
        (0, 110, 0),
        (0, 186, 171),
        (255, 145, 133),
        (133, 133, 0),
    ]



from pathlib import Path

from ..bliss_globals import current_session
from ..processor import BlissScanType
from ..utils.directories import get_filename


def export_filename_prefix(scan: BlissScanType, lima_name: str) -> str:
    scan_filename = Path(get_filename(scan))
    scan_nb = scan.scan_info.get("scan_nb")
    return f"{scan_filename.stem}_{scan_nb:04d}_{lima_name}"


def get_current_filename() -> str:
    return current_session.scan_saving.filename


def subtracted_nxprocess_name(lima_name: str) -> str:
    return f"{lima_name}_integrate_subtracted"



import logging
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from esrf_pathlib import ESRFPath
from ewoksutils.task_utils import task_inputs

from ..persistent.parameters import ParameterInfo
from ..processor import BlissScanType
from ..resources import resource_filename
from ..utils.directories import scan_processed_directory
from ..xrpd.processor import XrpdProcessor
from .utils import export_filename_prefix
from .utils import get_current_filename
from .utils import subtracted_nxprocess_name


class Bm02XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("config_filename", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
        ParameterInfo(
            "empty_cell_subtraction_options", category="Empty cell subtraction"
        ),
        ParameterInfo("ascii_export_enabled", category="ASCII export"),
        ParameterInfo("save_as_zip", category="ASCII export"),
        ParameterInfo("max_points", category="ASCII export"),
    ],
):

    DEFAULT_WORKFLOW: Optional[str] = resource_filename(
        "bm02", "integrate_scan_with_saving_subtract_ascii.json"
    )

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("config_filename", {"WOS": "", "D5": ""})
        defaults.setdefault(
            "integration_options",
            {
                "WOS": {
                    "method": "no_csr_cython",
                    "nbpt_rad": 4096,
                    "unit": "q_nm^-1",
                    "error_model": "poisson",
                },
                "D5": {
                    "method": "no_csr_cython",
                    "nbpt_rad": 4096,
                    "unit": "q_nm^-1",
                    "error_model": "poisson",
                },
            },
        )
        defaults.setdefault(
            "empty_cell_subtraction_options",
            {
                "CdTe": {
                    "enabled": False,
                    "cell_pattern_url": None,
                    "empty_cell_factor": "1",
                },
                "WOS": {
                    "enabled": False,
                    "cell_pattern_url": None,
                    "empty_cell_factor": 1,
                },
            },
        )

        defaults.setdefault("ascii_export_enabled", False)
        defaults.setdefault("save_as_zip", False)
        defaults.setdefault("max_points", 200)

        super().__init__(config=config, defaults=defaults)

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        try:
            return self.config_filename[lima_name]
        except KeyError:
            raise RuntimeError(
                f"Missing pyfai configuration file (poni or json) for '{lima_name}'"
            ) from None

    def get_integration_options(self, scan: BlissScanType, lima_name: str) -> dict:
        try:
            return self.integration_options.get(lima_name, {})
        except KeyError:
            raise RuntimeError(
                f"Missing pyfai integration options for '{lima_name}'"
            ) from None

    def _is_save_ascii_enabled(self, scan: BlissScanType, lima_name: str) -> bool:
        if not self.ascii_export_enabled:
            return False

        if "nbpt_azim" in self.get_integration_options(scan, lima_name):
            logging.warning(
                f'2D integrated patterns cannot be exported to ASCII. Disabling ASCII export for scan {scan.scan_info.get("scan_nb")}'
            )
            return False

        npoints = scan.scan_info.get("npoints", float("inf"))
        if npoints >= self.max_points:
            logging.warning(
                f"""
                Disabling ASCII export for scan {scan.scan_info.get("scan_nb")}
                 since its number of points ({npoints}) exceeds the max number ({self.max_points}).
                """
            )
            return False

        return True

    def enable_ascii_export(self):
        self.ascii_export_enabled = True

    def disable_ascii_export(self):
        self.ascii_export_enabled = False

    def _get_ascii_export_inputs(
        self, scan: BlissScanType, lima_name: str
    ) -> List[dict]:
        export_folder = Path(scan_processed_directory(scan)) / "export"
        filename_prefix = export_filename_prefix(scan, lima_name)
        ascii_basename_template = filename_prefix + "_%04d.dat"

        if self.save_as_zip:
            output_filename_template = ascii_basename_template
            output_archive_filename = export_folder / f"{filename_prefix}.zip"
        else:
            output_filename_template = export_folder / ascii_basename_template
            output_archive_filename = ""

        return task_inputs(
            task_identifier="SaveNexusPatternsAsAscii",
            inputs={
                "enabled": self._is_save_ascii_enabled(scan, lima_name),
                "output_filename_template": str(output_filename_template),
                "output_archive_filename": str(output_archive_filename),
            },
        )

    def _is_cell_subtraction_enabled(self, scan: BlissScanType, lima_name: str) -> bool:
        if "nbpt_azim" in self.get_integration_options(scan, lima_name):
            logging.warning(
                f'Cannot subtract empty cell from 2D integrated patterns. Disabling cell subtraction for scan {scan.scan_info.get("scan_nb")}'
            )
            return False

        if lima_name not in self.empty_cell_subtraction_options:
            return False

        return self.empty_cell_subtraction_options[lima_name].get("enabled", False)

    def _get_empty_cell_subtraction_inputs(
        self, scan: BlissScanType, lima_name: str
    ) -> List[dict]:
        inputs = self._get_data_access_inputs("SubtractBackgroundPattern")

        subtraction_options = self.empty_cell_subtraction_options.get(lima_name, {})

        inputs += task_inputs(
            task_identifier="SubtractBackgroundPattern",
            inputs={
                "enabled": self._is_cell_subtraction_enabled(scan, lima_name),
                "output_nxprocess_url": f"{self.master_output_url(scan)}/{subtracted_nxprocess_name(lima_name)}",
                "background_nxdata_url": subtraction_options.get(
                    "cell_pattern_url", None
                ),
                "background_factor": subtraction_options.get("empty_cell_factor", 1),
            },
        )

        external_output_url = self.external_output_url(scan, lima_name)
        if external_output_url:
            inputs.append(
                {
                    "task_identifier": "SubtractBackgroundPattern",
                    "name": "external_nxprocess_url",
                    "value": f"{external_output_url}/{subtracted_nxprocess_name(lima_name)}",
                }
            )

        return inputs

    def _data_to_plot_url(self, scan: BlissScanType, lima_name: str):
        if not scan.scan_info.get("save"):
            return None

        output_url = self.online_output_url(scan, lima_name)
        if self._is_cell_subtraction_enabled(scan, lima_name):
            return f"{output_url}/{subtracted_nxprocess_name(lima_name)}/integrated"
        else:
            return f"{output_url}/{lima_name}_integrate/integrated"

    def set_cell_pattern_url(
        self, sample: str, dataset: str, scan_number: int, detector: str
    ) -> None:

        filename = ESRFPath(get_current_filename())
        cell_filename = filename.replace_fields(
            collection=sample, dataset=dataset
        ).processed_dataset_file

        self.empty_cell_subtraction_options[detector][
            "cell_pattern_url"
        ] = f"{cell_filename}::/{scan_number}.1/{detector}_integrate/integrated"

    def enable_empty_cell_subtraction(self, *detectors):
        for detector_obj in detectors:
            detector: str = detector_obj.name
            if detector not in self.empty_cell_subtraction_options:
                self.empty_cell_subtraction_options[detector] = {}
            self.empty_cell_subtraction_options[detector]["enabled"] = True

            if (
                self.empty_cell_subtraction_options[detector].get("cell_pattern", None)
                is None
            ):
                print(
                    f"Do not forget to set a cell_pattern_url for {detector} now that cell subtraction is enabled!"
                )

    def disable_empty_cell_subtraction(self, *detectors):
        for detector in detectors:
            if detector.name not in self.empty_cell_subtraction_options:
                self.empty_cell_subtraction_options[detector.name] = {}
            self.empty_cell_subtraction_options[detector.name]["enabled"] = False

    def get_inputs(self, scan: BlissScanType, lima_name: str) -> List[dict]:
        inputs = super().get_inputs(scan, lima_name)
        inputs += self._get_ascii_export_inputs(scan, lima_name)
        inputs += self._get_empty_cell_subtraction_inputs(scan, lima_name)

        return inputs



"""Integration tests for the ID12 HDF5 converter."""

import pathlib

from ...bliss_globals import setup_globals
from .. import testing
from ..processors.id12 import DemoId12Hdf5ToSpecConverter

ID12_CONVERTER = DemoId12Hdf5ToSpecConverter()


@testing.integration_fixture
def id12_converter():
    ID12_CONVERTER.enable()
    yield ID12_CONVERTER
    ID12_CONVERTER.disable()


def id12_demo(expo=0.2, npoints=10):
    for _ in range(2):
        test_id12_loopscan(expo=expo, npoints=npoints)


@testing.integration_test
def test_id12_loopscan(
    id12_converter: DemoId12Hdf5ToSpecConverter, expo=0.2, npoints=10
):
    scan = setup_globals.loopscan(
        npoints,
        expo,
        setup_globals.diode1,
        setup_globals.diode2,
        setup_globals.mca1,
    )
    scan_number = scan.scan_info["scan_nb"]
    output_dir = pathlib.Path(id12_converter.output_dir(scan))
    _assert_spec_counter_file(id12_converter, output_dir, scan_number)


@testing.demo_assert("Check id12 SPEC file for scan #{scan_number}")
def _assert_spec_counter_file(
    id12_converter: DemoId12Hdf5ToSpecConverter,
    output_dir: pathlib.Path,
    scan_number: int,
):
    result = id12_converter._future.result(timeout=10)
    base_names = result["output_filenames"]
    assert len(base_names) == 1
    subscan = 1
    for base_name in base_names:
        with open(output_dir / base_name, "r") as f:
            assert f"scan{scan_number:03d}_{subscan}" in base_name, base_name
            counters = f.readline().split("  ")
            for name in [
                "epoch",
                "elapsed_time",
                "diode1",
                "diode2",
                "mca1_det0_events",
            ]:
                assert name in counters, name



from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional

from ..bliss_globals import current_session
from ..id31.xrpd_processor import Id31XrpdProcessor
from ..utils import directories
from ._id31_utils import ensure_difflab6_id31_flats
from .calib import DEFAULT_CALIB

_logger = logging.getLogger(__name__)


def ensure_demo_pyfai_config() -> str:
    """Create pyFAI config if not available yet and returns its filename"""
    processed_dir = directories.get_processed_dir(current_session.scan_saving.filename)
    config_path = Path(processed_dir, "config", "pyfaicalib.json")
    if not config_path.is_file():
        _logger.info(f"Create ID31 demo pyFAI config file: {str(config_path)}")
        config_path.parent.mkdir(parents=True, exist_ok=True)
        config_path.write_text(json.dumps(DEFAULT_CALIB))

    return str(config_path)


class DemoId31XrpdProcessor(Id31XrpdProcessor):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        if self._HAS_BLISS:
            pyfai_config = ensure_demo_pyfai_config()
            newflat, oldflat = ensure_difflab6_id31_flats()
            defaults.setdefault("pyfai_config", pyfai_config)

        defaults.setdefault("queue", "celery")
        defaults.setdefault("lima_names", ["difflab6"])

        super().__init__(config=config, defaults=defaults)

        if self._HAS_BLISS:
            self.newflat = newflat
            self.oldflat = oldflat


id31_xrpd_processor = DemoId31XrpdProcessor()



import pathlib
from typing import Tuple

import h5py
from silx.io.h5py_utils import retry
from silx.utils.retry import RetryError

from ._assert import demo_assert


@demo_assert("HDF5 dataset {data_path} in {filename} with shape {expected_shape}")
@retry(retry_timeout=10, retry_period=0.2)
def assert_hdf5_dataset_exists(
    filename: pathlib.Path,
    data_path: str,
    expected_shape: Tuple[int, ...],
) -> None:
    if not filename.exists():
        raise RetryError(f"{str(filename)!r} does not exist")
    filename = str(filename)
    with h5py.File(filename, mode="r") as root:
        if data_path not in root:
            raise RetryError(f"{data_path!r} not in {filename!r}")

        dset = root[data_path]
        if dset.shape != expected_shape:
            url = f"{filename!r}::{data_path}"
            raise RetryError(
                f"Shape of {url!r} is {dset.shape} instead of {expected_shape}"
            )



"""Integration tests for the generic XRPD processor."""

import os
import time

from ...bliss_globals import current_session
from ...bliss_globals import setup_globals
from ...xrpd.plots import Xrpd2dIntegrationPlot
from ...xrpd.plots import XrpdCurvePlot
from ...xrpd.plots import XrpdImagePlot
from .. import testing
from ..processors.xrpd import DemoXrpdProcessor

XRPD_PROCESSED = DemoXrpdProcessor()


def xrpd_demo_1d(nrepeats: int = None, expo=0.2, npoints=10):
    if nrepeats is None:
        nrepeats = 2

    for _ in range(nrepeats):
        test_xrpd_ct_with_1d_integration(expo=expo)
        test_xrpd_scan_with_1d_integration(expo=expo, npoints=npoints)


def xrpd_demo_2d(nrepeats: int = None, expo=0.2, npoints=10):
    if nrepeats is None:
        nrepeats = 2

    for _ in range(nrepeats):
        test_xrpd_ct_with_2d_integration(expo=expo)
        test_xrpd_scan_with_2d_integration(expo=expo, npoints=npoints)


def pct(*args, **kw):
    """``ct`` with workflow triggering"""
    s = setup_globals.ct(*args, **kw)
    XRPD_PROCESSED.on_new_scan(s)
    return s


@testing.integration_fixture
def xrpd_processor_1d():
    XRPD_PROCESSED.enable(setup_globals.difflab6)
    XRPD_PROCESSED._plotter.clear_lima_plots("difflab6")
    yield XRPD_PROCESSED
    XRPD_PROCESSED.disable()


@testing.integration_fixture
def xrpd_processor_2d():
    XRPD_PROCESSED.enable(setup_globals.difflab6)
    XRPD_PROCESSED.integration_options["nbpt_azim"] = 360
    XRPD_PROCESSED._plotter.clear_lima_plots("difflab6")
    yield XRPD_PROCESSED
    XRPD_PROCESSED.integration_options.pop("nbpt_azim", None)
    XRPD_PROCESSED.disable()


@testing.integration_test
def test_xrpd_ct_with_1d_integration(xrpd_processor_1d, expo=0.2):
    scan = pct(
        expo,
        setup_globals.difflab6,
        setup_globals.diode1,
        setup_globals.diode2,
    )
    testing.wait_workflows()
    _assert_1d_plot(scan, xrpd_processor_1d._plotter)
    _assert_1d_plot_last(scan, xrpd_processor_1d._plotter)


@testing.integration_test
def test_xrpd_scan_with_1d_integration(xrpd_processor_1d, expo=0.2, npoints=10):
    scan = setup_globals.loopscan(
        npoints,
        expo,
        setup_globals.difflab6,
        setup_globals.diode1,
        setup_globals.diode2,
    )
    testing.wait_workflows()
    _assert_1d_plot(scan, xrpd_processor_1d._plotter)
    _assert_1d_plot_last(scan, xrpd_processor_1d._plotter)
    _assert_1d_data(scan, npoints, xrpd_processor_1d)


@testing.integration_test
def test_xrpd_ct_with_2d_integration(xrpd_processor_2d, expo=0.2):
    scan = pct(
        expo,
        setup_globals.difflab6,
        setup_globals.diode1,
        setup_globals.diode2,
    )
    testing.wait_workflows()
    _assert_2d_plot_last(scan, xrpd_processor_2d._plotter)


@testing.integration_test
def test_xrpd_scan_with_2d_integration(xrpd_processor_2d, expo=0.2, npoints=10):
    scan = setup_globals.loopscan(
        npoints,
        expo,
        setup_globals.difflab6,
        setup_globals.diode1,
        setup_globals.diode2,
    )
    testing.wait_workflows()
    _assert_2d_plot_last(scan, xrpd_processor_2d._plotter)
    _assert_2d_data(scan, npoints, xrpd_processor_2d)


def _assert_1d_plot_last(scan, plotter):
    scan_legend = _get_scan_legend(scan)
    _assert_plot(plotter, scan_legend, "Integrated (Last)", XrpdCurvePlot)


def _assert_1d_plot(scan, plotter):
    scan_legend = _get_scan_legend(scan)
    _assert_plot(plotter, scan_legend, "Integrated difflab6", XrpdImagePlot)


def _assert_2d_plot_last(scan, plotter):
    scan_legend = _get_scan_legend(scan)
    _assert_plot(
        plotter, scan_legend, "2D Integrated difflab6 (last)", Xrpd2dIntegrationPlot
    )


@testing.demo_assert("Check {plot_id} plot for {scan_legend}")
def _assert_plot(
    plotter, scan_legend, plot_id, plot_cls, retry_timeout=10, retry_period=0.2
):
    image_plot = plotter._get_plot(plot_id, plot_cls)
    expected = [scan_legend]
    t0 = time.time()
    while True:
        labels = image_plot.get_labels()
        if labels == expected:
            return
        dt = time.time() - t0
        if dt > retry_timeout:
            break
        time.sleep(retry_period)
    assert labels == expected, labels


def _get_scan_legend(scan):
    filename = scan.scan_info["filename"]
    if not filename:
        filename = current_session.scan_saving.filename
    scan_nb = scan.scan_info["scan_nb"]
    scan_name = os.path.splitext(os.path.basename(filename))[0]
    return f"{scan_name}: {scan_nb}.1 {scan.name} (difflab6)"


@testing.demo_assert("Check 1D integration data")
def _assert_1d_data(scan, npoints, processor):
    data_keys = processor.get_data_keys(
        scan, "difflab6", retry_timeout=10, retry_period=0.2
    )

    shapes = {
        name: processor.get_data(scan, name, retry_timeout=10, retry_period=0.2).shape
        for name in data_keys
    }
    expected = {
        "difflab6:q": (4096,),
        "difflab6:intensity": (npoints, 4096),
        "difflab6:points": (npoints,),
    }
    assert shapes == expected, shapes


@testing.demo_assert("Check 2D integration data")
def _assert_2d_data(scan, npoints, processor):
    data_keys = processor.get_data_keys(
        scan, "difflab6", retry_timeout=10, retry_period=0.2
    )

    shapes = {
        name: processor.get_data(scan, name, retry_timeout=10, retry_period=0.2).shape
        for name in data_keys
    }
    expected = {
        "difflab6:chi": (360,),
        "difflab6:q": (4096,),
        "difflab6:intensity": (npoints, 360, 4096),
        "difflab6:points": (npoints,),
    }
    assert shapes == expected, shapes



from typing import List
from typing import Optional

from marshmallow import fields

from ..daiquiri.mixin import DaiquiriProcessorMixin
from ..daiquiri.mixin import ExposedParameter
from ..daiquiri.validators import exists_list
from ..daiquiri.validators import exists_valid_json
from ..persistent.parameters import ParameterInfo
from ..resources import resource_filename
from ..utils import directories
from .processor import BlissScanType
from .processor import FluoProcessor


class DaiquiriFluoProcessor(
    FluoProcessor,
    DaiquiriProcessorMixin,
    parameters=[ParameterInfo("notification_queue", category="daiquiri")],
):
    DEFAULT_WORKFLOW: Optional[str] = resource_filename(
        "fluo", "daiquiri_xrfmap_single_detector.json"
    )

    EXPOSED_PARAMETERS = [
        ExposedParameter(
            parameter="workflow",
            field_type=fields.Str,
            title="Current workflow",
            field_options={"validate": exists_valid_json},
        ),
        ExposedParameter(
            parameter="pymca_configs",
            field_type=fields.List,
            title="PyMCA Config File(s)",
            field_options={"validate": exists_list},
            field_args=[fields.Str()],
        ),
        ExposedParameter(
            parameter="quantification",
            field_type=fields.Bool,
            title="Quantification",
        ),
    ]

    def scan_requires_processing(self, scan) -> bool:
        """Only execute this workflow if there is a `daiquiri_datacollectionid` in the `scan_info`"""
        requires_processing = super().scan_requires_processing(scan) and bool(
            scan.scan_info.get("daiquiri_datacollectionid")
        )
        return requires_processing

    def get_daiquiri_inputs(self, scan: BlissScanType, workflow: str):
        task_identifier = "StartJob"
        return [
            {
                "task_identifier": task_identifier,
                "name": "dataCollectionId",
                "value": scan.scan_info.get("daiquiri_datacollectionid"),
            },
            {
                "task_identifier": task_identifier,
                "name": "output_filename",
                "value": directories.master_output_filename(scan),
            },
            {
                "task_identifier": task_identifier,
                "name": "workflow",
                "value": workflow,
            },
            {
                "task_identifier": task_identifier,
                "name": "config",
                "value": {
                    "pymca_configs": ",".join(self.pymca_configs),
                },
            },
            {
                "task_identifier": "NotifyBeamline",
                "name": "beamline",
                "value": self.notification_queue,
            },
        ]

    def get_workflow_and_inputs(self, scan: BlissScanType) -> List[dict]:
        """Merge in daiquiri StartJob options"""
        workflow, inputs = super().get_workflow_and_inputs(scan)
        inputs += self.get_daiquiri_inputs(scan, workflow)

        workflow = "daiquiri_" + workflow.replace(".ows", ".json")
        return workflow, inputs

    def get_submit_arguments(self, scan: BlissScanType, inputs: List[dict]) -> dict:
        """Need ppf for error handling"""
        args = super().get_submit_arguments(scan, inputs)
        args["engine"] = "ppf"
        return args



from blissoda.bliss_globals import setup_globals

from ..tests.itest_ewoks_macros import ewoks_macro_demo
from ..tests.itest_exafs import exafs_demo
from ..tests.itest_id12 import id12_demo
from ..tests.itest_id14 import id14_demo
from ..tests.itest_template import template_demo
from ..tests.itest_xrpd import xrpd_demo_1d
from ..tests.itest_xrpd import xrpd_demo_2d

# from ..user_scripts.bm02 import bm02_demo_1d
# from ..user_scripts.bm02 import bm02_demo_2d
# from ..user_scripts.id11 import id11_pdf_demo
# from ..user_scripts.id22 import id22_stscan_demo
# from ..user_scripts.id22 import id22_xrpd_demo
# from ..user_scripts.streamline import streamline_demo


def all_demo():
    print()
    print("===================")
    template_demo()

    # print()
    # print("===================")
    # setup_globals.newcollection("id22_stscan_collection")
    # id22_stscan_demo()

    # print()
    # print("===================")
    # setup_globals.newcollection("id22_xrpd_collection")
    # id22_xrpd_demo()

    # print()
    # print("===================")
    # setup_globals.newcollection("streamline_collection")
    # streamline_demo()

    # print()
    # print("===================")
    # setup_globals.newcollection("id11_collection")
    # print("TODO: get license for diffpy")
    # # id11_pdf_demo()

    print()
    print("===================")
    setup_globals.newcollection("xrpd_1d_collection")
    xrpd_demo_1d()

    print()
    print("===================")
    setup_globals.newcollection("xrpd_2d_collection")
    xrpd_demo_2d()

    # TODO: ewoksxrpd needs to be imported before xraylarch?
    #       Cannot load backend 'Qt5Agg' which requires the 'qt' interactive framework
    print()
    print("===================")
    setup_globals.newcollection("exafs_collection")
    exafs_demo()

    print()
    print("===================")
    setup_globals.newcollection("id14_collection")
    id14_demo()

    print()
    print("===================")
    setup_globals.newcollection("id12_collection")
    id12_demo()

    # print()
    # print("===================")
    # setup_globals.newcollection("bm02_collection")
    # bm02_demo_1d()
    # bm02_demo_2d()

    print()
    print("===================")
    setup_globals.newcollection("macro_collection")
    ewoks_macro_demo()

    print()
    print("===================")
    setup_globals.newcollection("done_collection")

    print()
    print("SUCCESS: all demos can be executed")



"""Define all fixtures that will be used for all tests."""

from ewoksjob.client.celery.utils import get_not_finished_futures

from ._display import print_message
from ._fixtures import integration_fixture


@integration_fixture(autouse=True)
def _cleanup_celery_jobs():
    yield
    futures = get_not_finished_futures()
    if futures:
        print_message(
            f"{len(futures)} workflows are still running: abort them.", "warning"
        )
    for future in futures:
        _ = future.abort()
        try:
            _ = future.result(timeout=5)
        except Exception as ex:
            print_message(f"Aborted workflow: {ex}", "warning")
